from __future__ import annotations

import collections
import math
import typing

import cairocffi

from libqtile import pangocffi, utils
from libqtile.log_utils import logger

if typing.TYPE_CHECKING:
    from libqtile.backend.base import Internal
    from libqtile.utils import ColorsType


class Drawer:
    """A helper class for drawing to Internal windows.

    We stage drawing operations locally in memory using a cairo RecordingSurface before
    finally drawing all operations to a backend-specific target.
    """

    def __init__(self, win: Internal, width: int, height: int):
        self._win = win
        self._width = width
        self._height = height

        self.surface: cairocffi.RecordingSurface
        self.last_surface: cairocffi.RecordingSurface
        self.ctx: cairocffi.Context
        self._reset_surface()

        self._has_mirrors = False

        self._enabled = True

    def finalize(self):
        """Destructor/Clean up resources"""
        if hasattr(self, "surface"):
            self.surface.finish()
            delattr(self, "surface")
        if hasattr(self, "last_surface"):
            self.last_surface.finish()
            delattr(self, "last_surface")
        self.ctx = None

    @property
    def has_mirrors(self):
        return self._has_mirrors

    @has_mirrors.setter
    def has_mirrors(self, value):
        if value and not self._has_mirrors:
            self._create_last_surface()

        self._has_mirrors = value

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, width: int):
        self._width = width

    @property
    def height(self) -> int:
        return self._height

    @height.setter
    def height(self, height: int):
        self._height = height

    def _reset_surface(self):
        """This creates a fresh surface and cairo context."""
        if hasattr(self, "surface"):
            self.surface.finish()

        self.surface = cairocffi.RecordingSurface(
            cairocffi.CONTENT_COLOR_ALPHA,
            None,
        )
        self.ctx = cairocffi.Context(self.surface)

    def _create_last_surface(self):
        """Creates a separate RecordingSurface for mirrors to access."""
        if hasattr(self, "last_surface"):
            self.last_surface.finish()
        self.last_surface = cairocffi.RecordingSurface(cairocffi.CONTENT_COLOR_ALPHA, None)

    def paint_to(self, drawer: Drawer) -> None:
        drawer.ctx.set_source_surface(self.last_surface)
        drawer.ctx.paint()

    def _rounded_rect(self, x, y, width, height, linewidth):
        aspect = 1.0
        corner_radius = height / 10.0
        radius = corner_radius / aspect
        degrees = math.pi / 180.0

        self.ctx.new_sub_path()

        delta = radius + linewidth / 2
        self.ctx.arc(x + width - delta, y + delta, radius, -90 * degrees, 0 * degrees)
        self.ctx.arc(x + width - delta, y + height - delta, radius, 0 * degrees, 90 * degrees)
        self.ctx.arc(x + delta, y + height - delta, radius, 90 * degrees, 180 * degrees)
        self.ctx.arc(x + delta, y + delta, radius, 180 * degrees, 270 * degrees)
        self.ctx.close_path()

    def rounded_rectangle(self, x: int, y: int, width: int, height: int, linewidth: int):
        self._rounded_rect(x, y, width, height, linewidth)
        self.ctx.set_line_width(linewidth)
        self.ctx.stroke()

    def rounded_fillrect(self, x: int, y: int, width: int, height: int, linewidth: int):
        self._rounded_rect(x, y, width, height, linewidth)
        self.ctx.fill()

    def rectangle(self, x: int, y: int, width: int, height: int, linewidth: int = 2):
        self.ctx.set_line_width(linewidth)
        self.ctx.rectangle(x, y, width, height)
        self.ctx.stroke()

    def fillrect(self, x: int, y: int, width: int, height: int, linewidth: int = 2):
        self.ctx.set_line_width(linewidth)
        self.ctx.rectangle(x, y, width, height)
        self.ctx.fill()
        self.ctx.stroke()

    def enable(self):
        """Enable drawing of surface to Internal window."""
        self._enabled = True

    def disable(self):
        """Disable drawing of surface to Internal window."""
        self._enabled = False

    def draw(
        self,
        offsetx: int = 0,
        offsety: int = 0,
        width: int | None = None,
        height: int | None = None,
        src_x: int = 0,
        src_y: int = 0,
    ):
        """
        A wrapper for the draw operation.

        This draws our cached operations to the Internal window.

        If Drawer has been disabled then the RecordingSurface will
        be cleared if no mirrors are waiting to copy its contents.

        Parameters
        ==========

        offsetx :
            the X offset to start drawing at.
        offsety :
            the Y offset to start drawing at.
        width :
            the X portion of the canvas to draw at the starting point.
        height :
            the Y portion of the canvas to draw at the starting point.
        src_x  :
            the X position of the origin in the source surface
        src_y  :
            the Y position of the origin in the source surface
        """
        if self._enabled:
            self._draw(
                offsetx=offsetx,
                offsety=offsety,
                width=width,
                height=height,
                src_x=src_x,
                src_y=src_y,
            )
            if self.has_mirrors:
                self._create_last_surface()
                ctx = cairocffi.Context(self.last_surface)
                ctx.set_source_surface(self.surface)
                ctx.paint()

        self._reset_surface()

    def _draw(
        self,
        offsetx: int = 0,
        offsety: int = 0,
        width: int | None = None,
        height: int | None = None,
        src_x: int = 0,
        src_y: int = 0,
    ):
        """
        This draws our cached operations to the Internal window.

        Parameters
        ==========

        offsetx :
            the X offset to start drawing at.
        offsety :
            the Y offset to start drawing at.
        width :
            the X portion of the canvas to draw at the starting point.
        height :
            the Y portion of the canvas to draw at the starting point.
        src_x  :
            the X position of the origin in the source surface
        src_y  :
            the Y position of the origin in the source surface
        """

    def set_source_rgb(self, colour: ColorsType, ctx: cairocffi.Context | None = None):
        # If an alternate context is not provided then we draw to the
        # drawer's default context
        if ctx is None:
            ctx = self.ctx
        if isinstance(colour, list):
            if len(colour) == 0:
                # defaults to black
                ctx.set_source_rgba(0.0, 0.0, 0.0, 1.0)
            elif len(colour) == 1:
                ctx.set_source_rgba(*utils.rgb(colour[0]))
            else:
                linear = cairocffi.LinearGradient(0.0, 0.0, 0.0, self.height)
                step_size = 1.0 / (len(colour) - 1)
                step = 0.0
                for c in colour:
                    linear.add_color_stop_rgba(step, *utils.rgb(c))
                    step += step_size
                ctx.set_source(linear)
        else:
            ctx.set_source_rgba(*utils.rgb(colour))

    def clear_rect(self, x=0, y=0, width=0, height=0):
        """
        Erases the background area specified by parameters. By default,
        the whole Drawer is cleared.

        The ability to clear a smaller area may be useful when you want to
        erase a smaller area of the drawer (e.g. drawing widget decorations).
        """
        if width <= 0:
            width = self.width
        if height <= 0:
            height = self.height

        self.ctx.save()
        self.ctx.set_operator(cairocffi.OPERATOR_CLEAR)
        self.ctx.rectangle(x, y, width, height)
        self.ctx.fill()
        self.ctx.restore()

    def clear(self, colour):
        """Clears background of the Drawer and fills with specified colour."""
        if self.ctx is None:
            self._reset_surface()
        self.ctx.save()

        # Erase the background
        self.clear_rect()

        # Fill drawer with new colour
        self.ctx.set_operator(cairocffi.OPERATOR_SOURCE)
        self.set_source_rgb(colour)
        self.ctx.rectangle(0, 0, self.width, self.height)
        self.ctx.fill()

        self.ctx.restore()

    def textlayout(self, text, colour, font_family, font_size, font_shadow, markup=False, **kw):
        """Get a text layout"""
        textlayout = TextLayout(
            self, text, colour, font_family, font_size, font_shadow, markup=markup, **kw
        )
        return textlayout

    def max_layout_size(self, texts, font_family, font_size, markup=False):
        sizelayout = self.textlayout("", "ffffff", font_family, font_size, None, markup=markup)
        widths, heights = [], []
        for i in texts:
            sizelayout.text = i
            widths.append(sizelayout.width)
            heights.append(sizelayout.height)
        return max(widths), max(heights)

    def text_extents(self, text):
        return self.ctx.text_extents(utils.scrub_to_utf8(text))

    def font_extents(self):
        return self.ctx.font_extents()

    def fit_fontsize(self, heightlimit):
        """Try to find a maximum font size that fits any strings within the height"""
        self.ctx.set_font_size(heightlimit)
        asc, desc, height, _, _ = self.font_extents()
        self.ctx.set_font_size(int(heightlimit * heightlimit / height))
        return self.font_extents()

    def fit_text(self, strings, heightlimit):
        """Try to find a maximum font size that fits all strings within the height"""
        self.ctx.set_font_size(heightlimit)
        _, _, _, maxheight, _, _ = self.ctx.text_extents("".join(strings))
        if not maxheight:
            return 0, 0
        self.ctx.set_font_size(int(heightlimit * heightlimit / maxheight))
        maxwidth, maxheight = 0, 0
        for i in strings:
            _, _, x, y, _, _ = self.ctx.text_extents(i)
            maxwidth = max(maxwidth, x)
            maxheight = max(maxheight, y)
        return maxwidth, maxheight

    def draw_vbar(self, color, x, y1, y2, linewidth=1):
        self.set_source_rgb(color)
        self.ctx.move_to(x, y1)
        self.ctx.line_to(x, y2)
        self.ctx.set_line_width(linewidth)
        self.ctx.stroke()

    def draw_hbar(self, color, x1, x2, y, linewidth=1):
        self.set_source_rgb(color)
        self.ctx.move_to(x1, y)
        self.ctx.line_to(x2, y)
        self.ctx.set_line_width(linewidth)
        self.ctx.stroke()


class TextLayout:
    def __init__(
        self, drawer, text, colour, font_family, font_size, font_shadow, wrap=True, markup=False
    ):
        self.drawer, self.colour = drawer, colour
        layout = pangocffi.create_layout(drawer.ctx)
        layout.set_alignment(pangocffi.ALIGN_CENTER)
        if not wrap:  # pango wraps by default
            layout.set_ellipsize(pangocffi.ELLIPSIZE_END)
        desc = pangocffi.FontDescription.from_string(f"{font_family} {font_size}px")
        layout.set_font_description(desc)
        self.font_shadow = font_shadow
        self.layout = layout
        self.markup = markup
        self.text = text
        self._width = None

    def finalize(self):
        self.layout.finalize()

    @property
    def text(self):
        return self.layout.get_text()

    @text.setter
    def text(self, value):
        if self.markup:
            # pangocffi doesn't like None here, so we use "".
            if value is None:
                value = ""
            try:
                attrlist, value, accel_char = pangocffi.parse_markup(value)
                self.layout.set_attributes(attrlist)
            except pangocffi.BadMarkup as e:
                logger.warning(e)
        self.layout.set_text(utils.scrub_to_utf8(value))

    @property
    def width(self):
        if self._width is not None:
            return self._width
        else:
            return self.layout.get_pixel_size()[0]

    @width.setter
    def width(self, value):
        self._width = value
        self.layout.set_width(pangocffi.units_from_double(value))

    def reset_width(self):
        self._width = None
        self.layout.set_width(-1)

    @property
    def height(self):
        return self.layout.get_pixel_size()[1]

    def fontdescription(self):
        return self.layout.get_font_description()

    @property
    def font_family(self):
        d = self.fontdescription()
        return d.get_family()

    @font_family.setter
    def font_family(self, font):
        d = self.fontdescription()
        d.set_family(font)
        self.layout.set_font_description(d)

    @property
    def font_size(self):
        d = self.fontdescription()
        return pangocffi.units_to_double(int(d.get_size()))

    @font_size.setter
    def font_size(self, size):
        d = self.fontdescription()
        d.set_absolute_size(pangocffi.units_from_double(size))
        self.layout.set_font_description(d)

    def draw(self, x, y):
        if self.font_shadow is not None:
            self.drawer.set_source_rgb(self.font_shadow)
            self.drawer.ctx.move_to(x + 1, y + 1)
            pangocffi.show_layout(self.drawer.ctx, self.layout)

        self.drawer.set_source_rgb(self.colour)
        self.drawer.ctx.move_to(x, y)
        pangocffi.show_layout(self.drawer.ctx, self.layout)

    def framed(self, border_width, border_color, pad_x, pad_y, highlight_color=None):
        return TextFrame(
            self, border_width, border_color, pad_x, pad_y, highlight_color=highlight_color
        )


class TextFrame:
    def __init__(self, layout, border_width, border_color, pad_x, pad_y, highlight_color=None):
        self.layout = layout
        self.border_width = border_width
        self.border_color = border_color
        self.drawer = self.layout.drawer
        self.highlight_color = highlight_color

        if isinstance(pad_x, collections.abc.Iterable):
            self.pad_left = pad_x[0]
            self.pad_right = pad_x[1]
        else:
            self.pad_left = self.pad_right = pad_x

        if isinstance(pad_y, collections.abc.Iterable):
            self.pad_top = pad_y[0]
            self.pad_bottom = pad_y[1]
        else:
            self.pad_top = self.pad_bottom = pad_y

    def draw(self, x, y, rounded=True, fill=False, line=False, highlight=False):
        self.drawer.set_source_rgb(self.border_color)
        opts = [
            x,
            y,
            self.layout.width + self.pad_left + self.pad_right,
            self.layout.height + self.pad_top + self.pad_bottom,
            self.border_width,
        ]
        if line:
            if highlight:
                self.drawer.set_source_rgb(self.highlight_color)
                self.drawer.fillrect(*opts)
                self.drawer.set_source_rgb(self.border_color)

            # change to only fill in bottom line
            opts[1] = self.height - self.border_width  # y
            opts[3] = self.border_width  # height

            self.drawer.fillrect(*opts)
        elif fill:
            if rounded:
                self.drawer.rounded_fillrect(*opts)
            else:
                self.drawer.fillrect(*opts)
        else:
            if rounded:
                self.drawer.rounded_rectangle(*opts)
            else:
                self.drawer.rectangle(*opts)
        self.drawer.ctx.stroke()
        self.layout.draw(x + self.pad_left, y + self.pad_top)

    def draw_fill(self, x, y, rounded=True):
        self.draw(x, y, rounded=rounded, fill=True)

    def draw_line(self, x, y, highlighted):
        self.draw(x, y, line=True, highlight=highlighted)

    @property
    def height(self):
        return self.layout.height + self.pad_top + self.pad_bottom

    @property
    def width(self):
        return self.layout.width + self.pad_left + self.pad_right



import os
from collections.abc import Callable

import libqtile.resources
from libqtile import bar, layout, qtile, widget
from libqtile.config import Click, Drag, Group, Key, Match, Output, Screen
from libqtile.lazy import lazy
from libqtile.utils import guess_terminal

mod = "mod4"
terminal = guess_terminal()

keys = [
    # A list of available commands that can be bound to keys can be found
    # at https://docs.qtile.org/en/latest/manual/config/lazy.html
    # Switch between windows
    Key([mod], "h", lazy.layout.left(), desc="Move focus to left"),
    Key([mod], "l", lazy.layout.right(), desc="Move focus to right"),
    Key([mod], "j", lazy.layout.down(), desc="Move focus down"),
    Key([mod], "k", lazy.layout.up(), desc="Move focus up"),
    Key([mod], "space", lazy.layout.next(), desc="Move window focus to other window"),
    # Move windows between left/right columns or move up/down in current stack.
    # Moving out of range in Columns layout will create new column.
    Key([mod, "shift"], "h", lazy.layout.shuffle_left(), desc="Move window to the left"),
    Key([mod, "shift"], "l", lazy.layout.shuffle_right(), desc="Move window to the right"),
    Key([mod, "shift"], "j", lazy.layout.shuffle_down(), desc="Move window down"),
    Key([mod, "shift"], "k", lazy.layout.shuffle_up(), desc="Move window up"),
    # Grow windows. If current window is on the edge of screen and direction
    # will be to screen edge - window would shrink.
    Key([mod, "control"], "h", lazy.layout.grow_left(), desc="Grow window to the left"),
    Key([mod, "control"], "l", lazy.layout.grow_right(), desc="Grow window to the right"),
    Key([mod, "control"], "j", lazy.layout.grow_down(), desc="Grow window down"),
    Key([mod, "control"], "k", lazy.layout.grow_up(), desc="Grow window up"),
    Key([mod], "n", lazy.layout.normalize(), desc="Reset all window sizes"),
    # Toggle between split and unsplit sides of stack.
    # Split = all windows displayed
    # Unsplit = 1 window displayed, like Max layout, but still with
    # multiple stack panes
    Key(
        [mod, "shift"],
        "Return",
        lazy.layout.toggle_split(),
        desc="Toggle between split and unsplit sides of stack",
    ),
    Key([mod], "Return", lazy.spawn(terminal), desc="Launch terminal"),
    # Toggle between different layouts as defined below
    Key([mod], "Tab", lazy.next_layout(), desc="Toggle between layouts"),
    Key([mod], "w", lazy.window.kill(), desc="Kill focused window"),
    Key(
        [mod],
        "f",
        lazy.window.toggle_fullscreen(),
        desc="Toggle fullscreen on the focused window",
    ),
    Key([mod], "t", lazy.window.toggle_floating(), desc="Toggle floating on the focused window"),
    Key([mod, "control"], "r", lazy.reload_config(), desc="Reload the config"),
    Key([mod, "control"], "q", lazy.shutdown(), desc="Shutdown Qtile"),
    Key([mod], "r", lazy.spawncmd(), desc="Spawn a command using a prompt widget"),
]

# Add key bindings to switch VTs in Wayland.
# We can't check qtile.core.name in default config as it is loaded before qtile is started
# We therefore defer the check until the key binding is run by using .when(func=...)
for vt in range(1, 8):
    keys.append(
        Key(
            ["control", "mod1"],
            f"f{vt}",
            lazy.core.change_vt(vt).when(func=lambda: qtile.core.name == "wayland"),
            desc=f"Switch to VT{vt}",
        )
    )


groups = [Group(i) for i in "123456789"]

for i in groups:
    keys.extend(
        [
            # mod + group number = switch to group
            Key(
                [mod],
                i.name,
                lazy.group[i.name].toscreen(),
                desc=f"Switch to group {i.name}",
            ),
            # mod + shift + group number = switch to & move focused window to group
            Key(
                [mod, "shift"],
                i.name,
                lazy.window.togroup(i.name, switch_group=True),
                desc=f"Switch to & move focused window to group {i.name}",
            ),
            # Or, use below if you prefer not to switch to that group.
            # # mod + shift + group number = move focused window to group
            # Key([mod, "shift"], i.name, lazy.window.togroup(i.name),
            #     desc="move focused window to group {}".format(i.name)),
        ]
    )

layouts = [
    layout.Columns(border_focus_stack=["#d75f5f", "#8f3d3d"], border_width=4),
    layout.Max(),
    # Try more layouts by unleashing below layouts.
    # layout.Stack(num_stacks=2),
    # layout.Bsp(),
    # layout.Matrix(),
    # layout.MonadTall(),
    # layout.MonadWide(),
    # layout.RatioTile(),
    # layout.Tile(),
    # layout.TreeTab(),
    # layout.VerticalTile(),
    # layout.Zoomy(),
]

widget_defaults = dict(
    font="sans",
    fontsize=12,
    padding=3,
)
extension_defaults = widget_defaults.copy()

logo = os.path.join(os.path.dirname(libqtile.resources.__file__), "logo.png")
screens = [
    Screen(
        bottom=bar.Bar(
            [
                widget.CurrentLayout(),
                widget.GroupBox(),
                widget.Prompt(),
                widget.WindowName(),
                widget.Chord(
                    chords_colors={
                        "launch": ("#ff0000", "#ffffff"),
                    },
                    name_transform=lambda name: name.upper(),
                ),
                widget.TextBox("default config", name="default"),
                widget.TextBox("Press &lt;M-r&gt; to spawn", foreground="#d75f5f"),
                # NB Systray is incompatible with Wayland, consider using StatusNotifier instead
                # widget.StatusNotifier(),
                widget.Systray(),
                widget.Clock(format="%Y-%m-%d %a %I:%M %p"),
                widget.QuickExit(),
            ],
            24,
            # border_width=[2, 0, 2, 0],  # Draw top and bottom borders
            # border_color=["ff00ff", "000000", "ff00ff", "000000"]  # Borders are magenta
        ),
        background="#000000",
        wallpaper=logo,
        wallpaper_mode="center",
        # You can uncomment this variable if you see that on X11 floating resize/moving is laggy
        # By default we handle these events delayed to already improve performance, however your system might still be struggling
        # This variable is set to None (no cap) by default, but you can set it to 60 to indicate that you limit it to 60 events per second
        # x11_drag_polling_rate = 60,
    ),
]

# Instead of screens, you can define a function here to specify which Screen
# should correspond to which Output.
fake_screens: list[Screen] | None = None

# Instead of screens or fake screens, you can define a function here that
# returns a list of Screen objects based on the list of Outputs; that way you
# can decide based on e.g. the number of screens, or which ports are plugged
# in exactly what do render in each bar for each screen.
generate_screens: Callable[[list[Output]], list[Screen]] | None = None

# Drag floating layouts.
mouse = [
    Drag([mod], "Button1", lazy.window.set_position_floating(), start=lazy.window.get_position()),
    Drag([mod], "Button3", lazy.window.set_size_floating(), start=lazy.window.get_size()),
    Click([mod], "Button2", lazy.window.bring_to_front()),
]

dgroups_key_binder = None
dgroups_app_rules = []  # type: list
follow_mouse_focus = True
bring_front_click = False
floats_kept_above = True
cursor_warp = False
floating_layout = layout.Floating(
    float_rules=[
        # Run the utility of `xprop` to see the wm class and name of an X client.
        *layout.Floating.default_float_rules,
        Match(wm_class="confirmreset"),  # gitk
        Match(wm_class="makebranch"),  # gitk
        Match(wm_class="maketag"),  # gitk
        Match(wm_class="ssh-askpass"),  # ssh-askpass
        Match(title="branchdialog"),  # gitk
        Match(title="pinentry"),  # GPG key password entry
    ]
)
auto_fullscreen = True
focus_on_window_activation = "smart"
focus_previous_on_window_remove = False
reconfigure_screens = True

# If things like steam games want to auto-minimize themselves when losing
# focus, should we respect this or not?
auto_minimize = True

# When using the Wayland backend, this can be used to configure input devices.
wl_input_rules = None

# xcursor theme (string or None) and size (integer) for Wayland backend
wl_xcursor_theme = None
wl_xcursor_size = 24

idle_timers = []  # type: list
idle_inhibitors = []  # type: list

# XXX: Gasp! We're lying here. In fact, nobody really uses or cares about this
# string besides java UI toolkits; you can see several discussions on the
# mailing lists, GitHub issues, and other WM documentation that suggest setting
# this string if your java app doesn't work correctly. We may as well just lie
# and say that we're a working one by default.
#
# We choose LG3D to maximize irony: it is a 3D non-reparenting WM written in
# java that happens to be on java's whitelist.
wmname = "LG3D"



import re
import subprocess

from libqtile import bar
from libqtile.widget import base


class TunedManager(base.BackgroundPoll):
    """
    A widget to interact with the Tuned power management daemon.
    It always displays the name of the currently active profile.

    The user can define a list of profiles to be used, 3 default
    profiles exist. These 3 are the default profiles on RHEL and
    Fedora Linux.

    A left click on the widget will go to the next layout in the
    list, a right click will go to the previous one.  If the end
    of the list is reached, it cycles back to the beginning.
    Scrolling can also be used to cycle through the list, though
    keep in mind that switching the profile takes a while and so
    scrolling through the list quickly is not feasible.
    """

    orientations = base.ORIENTATION_HORIZONTAL

    defaults = [
        (
            "modes",
            ["powersave", "balanced-battery", "throughput-performance"],
            "The modes to cycle through",
        )
    ]

    def __init__(self, **config):
        base.BackgroundPoll.__init__(self, "", **config)
        self.add_defaults(TunedManager.defaults)
        self.length_type = bar.CALCULATED
        self.regex = re.compile(r"Current active profile:\s+(\S+)")
        self.current_mode = self.find_mode()

        self.add_callbacks(
            {
                "Button1": self.next_mode,  # Left Click
                "Button3": self.previous_mode,  # Right Click
                "Button4": self.next_mode,  # Scroll up
                "Button5": self.previous_mode,  # Scroll down
            }
        )

    def _configure(self, qtile, bar):
        base.BackgroundPoll._configure(self, qtile, bar)
        self.text = self.find_mode()

    def poll(self):
        return self.find_mode()

    def find_mode(self):
        result = subprocess.run("tuned-adm active", shell=True, capture_output=True, text=True)
        output = result.stdout
        mode = self.regex.findall(output)
        if not mode:
            return ""
        return mode[0]

    def update_bar(self):
        self.current_mode = self.find_mode()
        self.text = self.current_mode
        self.bar.draw()

    def execute_command(self, index: int):
        argument = self.modes[index]  # pyright: ignore
        try:
            subprocess.run(["tuned-adm", "profile", argument], check=True)
            self.update_bar()
        except subprocess.CalledProcessError as e:
            self.update(f"Error setting mode: {e}")

    def _change_mode(self, step=1):
        next_index: int = (self.modes.index(self.current_mode) + step) % len(self.modes)  # pyright: ignore
        self.execute_command(next_index)

    def next_mode(self):
        self._change_mode()

    def previous_mode(self):
        self._change_mode(step=-1)



import contextlib
import copy
import time
from enum import Enum, Flag, auto
from math import isclose
from typing import NamedTuple

from libqtile import hook
from libqtile.backend.base import Window
from libqtile.command.base import expose_command
from libqtile.hook import Hook, qtile_hooks
from libqtile.layout.base import Layout

plasma_hook = Hook(
    "plasma_add_mode",
    """
    Used to flag when the add mode of the Plasma layout has changed.

    The hooked function should take one argument being the layout object.
    """,
)


qtile_hooks.register_hook(plasma_hook)


class NotRestorableError(Exception):
    pass


class Point(NamedTuple):
    x: int
    y: int


class Dimensions(NamedTuple):
    x: int
    y: int
    width: int
    height: int


class Orient(Flag):
    HORIZONTAL = 0
    VERTICAL = 1


class Direction(Enum):
    UP = auto()
    DOWN = auto()
    LEFT = auto()
    RIGHT = auto()

    @property
    def orient(self):
        return Orient.HORIZONTAL if self in [self.LEFT, self.RIGHT] else Orient.VERTICAL

    @property
    def offset(self):
        return 1 if self in [self.RIGHT, self.DOWN] else -1


class Priority(Enum):
    FIXED = auto()
    BALANCED = auto()


class AddMode(Flag):
    HORIZONTAL = auto()
    VERTICAL = auto()
    SPLIT = auto()

    @property
    def orient(self):
        return Orient.VERTICAL if self & self.VERTICAL else Orient.HORIZONTAL


border_check = {
    Direction.UP: lambda a, b: isclose(a.y, b.y_end),
    Direction.DOWN: lambda a, b: isclose(a.y_end, b.y),
    Direction.LEFT: lambda a, b: isclose(a.x, b.x_end),
    Direction.RIGHT: lambda a, b: isclose(a.x_end, b.x),
}


def flatten(value):
    """Flattens a nested list of lists into a single list."""
    out = []
    for x in value:
        if not isinstance(x, list):
            out.append(x)
        else:
            out.extend(flatten(x))
    return out


class Node:
    """
    A tree node.

    Each node represents a container that can hold a payload and child nodes.
    """

    min_size_default = 100
    root_orient = Orient.HORIZONTAL
    priority = Priority.FIXED

    def __init__(self, payload=None, x=None, y=None, width=None, height=None):
        self.payload = payload
        self._x = x
        self._y = y
        self._width = width
        self._height = height
        self._size = None
        self.children = []
        self.last_accessed = 0
        self.parent = None
        self.restorables = {}
        hook.subscribe.client_killed(self._clean_restorables)

    def __repr__(self):
        info = self.payload or ""
        if self:
            info += f" +{len(self):d}"
        return f"<Node {info} {id(self):x}>"

    # Define dunder methods to treat Node objects like an iterable
    def __contains__(self, node):
        if node is self:
            return True
        for child in self:
            if node in child:
                return True
        return False

    def __iter__(self):
        yield from self.children

    def __getitem__(self, key):
        return self.children[key]

    def __setitem__(self, key, value):
        self.children[key] = value

    def __len__(self):
        return len(self.children)

    @property
    def root(self):
        try:
            # Walk way up the tree until we find the root
            return self.parent.root
        except AttributeError:
            # Node has no parent (self.parent is None) so node must be root
            return self

    @property
    def is_root(self):
        return self.parent is None

    @property
    def is_leaf(self):
        return not self.children

    @property
    def index(self):
        return self.parent.children.index(self)

    @property
    def tree(self):
        return [c.tree if c else c for c in self]

    @property
    def siblings(self):
        if self.is_root:
            return list()
        return [c for c in self.parent if c is not self]

    @property
    def first_leaf(self):
        if self.is_leaf:
            return self
        return self[0].first_leaf

    @property
    def last_leaf(self):
        if self.is_leaf:
            return self
        return self[-1].last_leaf

    @property
    def recent_leaf(self):
        if self.is_leaf:
            return self
        return max(self, key=lambda n: n.last_accessed).recent_leaf

    @property
    def prev_leaf(self):
        if self.is_root:
            return self.last_leaf
        idx = self.index - 1
        if idx < 0:
            return self.parent.prev_leaf
        return self.parent[idx].last_leaf

    @property
    def next_leaf(self):
        if self.is_root:
            return self.first_leaf
        idx = self.index + 1
        if idx >= len(self.parent):
            return self.parent.next_leaf
        return self.parent[idx].first_leaf

    @property
    def all_leafs(self):
        if self.is_leaf:
            yield self
        for child in self:
            yield from child.all_leafs

    @property
    def orient(self):
        if self.is_root:
            return self.root_orient
        return ~self.parent.orient

    @property
    def horizontal(self):
        return self.orient is Orient.HORIZONTAL

    @property
    def vertical(self):
        return self.orient is Orient.VERTICAL

    @property
    def x(self):
        if self.is_root:
            return self._x
        if self.horizontal:
            return self.parent.x
        return self.parent.x + self.size_offset

    @x.setter
    def x(self, val):
        if not self.is_root:
            return
        self._x = val

    @property
    def y(self):
        if self.is_root:
            return self._y
        if self.vertical:
            return self.parent.y
        return self.parent.y + self.size_offset

    @y.setter
    def y(self, val):
        if not self.is_root:
            return
        self._y = val

    @property
    def pos(self):
        return Point(self.x, self.y)

    @property
    def width(self):
        if self.is_root:
            return self._width
        if self.horizontal:
            return self.parent.width
        return self.size

    @width.setter
    def width(self, val):
        if self.is_root:
            self._width = val
        elif self.horizontal:
            self.parent.size = val
        else:
            self.size = val

    @property
    def height(self):
        if self.is_root:
            return self._height
        if self.vertical:
            return self.parent.height
        return self.size

    @height.setter
    def height(self, val):
        if self.is_root:
            self._height = val
        elif self.vertical:
            self.parent.size = val
        else:
            self.size = val

    @property
    def x_end(self):
        return self.x + self.width

    @property
    def y_end(self):
        return self.y + self.height

    @property
    def x_center(self):
        return self.x + self.width / 2

    @property
    def y_center(self):
        return self.y + self.height / 2

    @property
    def center(self):
        return Point(self.x_center, self.y_center)

    @property
    def pixel_perfect(self):
        """
        Return pixel-perfect int dimensions (x, y, width, height) which
        compensate for gaps in the layout grid caused by plain int conversions.
        """
        x, y, width, height = self.x, self.y, self.width, self.height
        threshold = 0.99999
        if (x - int(x)) + (width - int(width)) > threshold:
            width += 1
        if (y - int(y)) + (height - int(height)) > threshold:
            height += 1
        return Dimensions(*map(int, (x, y, width, height)))

    @property
    def capacity(self):
        return self.width if self.horizontal else self.height

    @property
    def size(self):
        """Return amount of space taken in parent container."""
        if self.is_root:
            return None
        if self.fixed:
            return self._size
        if self.flexible:
            # Distribute space evenly among flexible nodes
            taken = sum(n.size for n in self.siblings if not n.flexible)
            flexibles = [n for n in self.parent if n.flexible]
            return (self.parent.capacity - taken) / len(flexibles)
        return max(sum(gc.min_size for gc in c) for c in self)

    @size.setter
    def size(self, val):
        if self.is_root or not self.siblings:
            return
        if val is None:
            self.reset_size()
            return
        occupied = sum(s.min_size_bound for s in self.siblings)
        val = max(min(val, self.parent.capacity - occupied), self.min_size_bound)
        self.force_size(val)

    def force_size(self, val):
        """Set size without considering available space."""
        Node.fit_into(self.siblings, self.parent.capacity - val)
        if val == 0:
            return
        if self:
            Node.fit_into([self], val)
        self._size = val

    @property
    def size_offset(self):
        return sum(c.size for c in self.parent[: self.index])

    @staticmethod
    def fit_into(nodes, space):
        """Resize nodes to fit them into the available space."""
        if not nodes:
            return
        occupied = sum(n.min_size for n in nodes)
        if space >= occupied and any(n.flexible for n in nodes):
            # If any flexible node exists, it will occupy the space
            # automatically, not requiring any action.
            return
        nodes_left = nodes[:]
        space_left = space
        if space < occupied:
            for node in nodes:
                if node.min_size_bound != node.min_size:
                    continue
                # Substract nodes that are already at their minimal possible
                # size because they can't be shrinked any further.
                space_left -= node.min_size
                nodes_left.remove(node)
        if not nodes_left:
            return
        factor = space_left / sum(n.size for n in nodes_left)
        for node in nodes_left:
            new_size = node.size * factor
            if node.fixed:
                node._size = new_size  # pylint: disable=protected-access
            for child in node:
                Node.fit_into(child, new_size)

    @property
    def fixed(self):
        """A node is fixed if it has a specified size."""
        return self._size is not None

    @property
    def min_size(self):
        if self.fixed:
            return self._size
        if self.is_leaf:
            return self.min_size_default
        size = max(sum(gc.min_size for gc in c) for c in self)
        return max(size, self.min_size_default)

    @property
    def min_size_bound(self):
        if self.is_leaf:
            return self.min_size_default
        return max(sum(gc.min_size_bound for gc in c) or self.min_size_default for c in self)

    def reset_size(self):
        self._size = None

    @property
    def flexible(self):
        """
        A node is flexible if its size isn't (explicitly or implicitly)
        determined.
        """
        if self.fixed:
            return False
        return all((any(gc.flexible for gc in c) or c.is_leaf) for c in self)

    def access(self):
        self.last_accessed = time.time()
        try:
            self.parent.access()
        except AttributeError:
            pass

    def neighbor(self, direction):
        """Return adjacent leaf node in specified direction."""
        if self.is_root:
            return None
        if direction.orient is self.parent.orient:
            target_idx = self.index + direction.offset
            if 0 <= target_idx < len(self.parent):
                return self.parent[target_idx].recent_leaf
            if self.parent.is_root:
                return None
            return self.parent.parent.neighbor(direction)
        return self.parent.neighbor(direction)

    @property
    def up(self):
        return self.neighbor(Direction.UP)

    @property
    def down(self):
        return self.neighbor(Direction.DOWN)

    @property
    def left(self):
        return self.neighbor(Direction.LEFT)

    @property
    def right(self):
        return self.neighbor(Direction.RIGHT)

    def common_border(self, node, direction):
        """Return whether a common border with given node in specified
        direction exists.
        """
        if not border_check[direction](self, node):
            return False
        if direction in [Direction.UP, Direction.DOWN]:
            detached = node.x >= self.x_end or node.x_end <= self.x
        else:
            detached = node.y >= self.y_end or node.y_end <= self.y
        return not detached

    def close_neighbor(self, direction):
        """Return visually adjacent leaf node in specified direction."""
        nodes = [n for n in self.root.all_leafs if self.common_border(n, direction)]
        if not nodes:
            return None
        most_recent = max(nodes, key=lambda n: n.last_accessed)
        if most_recent.last_accessed > 0:
            return most_recent
        if direction in [Direction.UP, Direction.DOWN]:
            match = lambda n: n.x <= self.x_center <= n.x_end  # noqa: E731
        else:
            match = lambda n: n.y <= self.y_center <= n.y_end  # noqa: E731
        return next(n for n in nodes if match(n))

    @property
    def close_up(self):
        return self.close_neighbor(Direction.UP)

    @property
    def close_down(self):
        return self.close_neighbor(Direction.DOWN)

    @property
    def close_left(self):
        return self.close_neighbor(Direction.LEFT)

    @property
    def close_right(self):
        return self.close_neighbor(Direction.RIGHT)

    def add_child(self, node, idx=None):
        if idx is None:
            idx = len(self)
        self.children.insert(idx, node)
        node.parent = self
        if len(self) == 1:
            return
        total = self.capacity
        if Node.priority is Priority.FIXED:
            # Prioritising windows with fixed sizes means the most space the siblings
            # must fit into is total width less the minimum size for a new node.
            # However, the new node doesn't have a fixed size so will expand to fit
            # available space
            space = total - Node.min_size_default
        else:
            # Balanced approach means that space for existing nodes is reduced so that
            # all nodes would be distributed evenly if none had fixed widths
            space = total - (total / len(self))
        Node.fit_into(node.siblings, space)

    def add_child_after(self, new, old):
        self.add_child(new, idx=old.index + 1)

    def remove_child(self, node):
        node._save_restore_state()  # pylint: disable=W0212
        node.force_size(0)
        self.children.remove(node)
        if len(self) == 1:
            child = self[0]
            if self.is_root:
                # A single child doesn't need a fixed size
                child.reset_size()
            else:
                # Collapse tree with a single child
                self.parent.replace_child(self, child)
                Node.fit_into(child, self.capacity)

    def remove(self):
        self.parent.remove_child(self)

    def replace_child(self, old, new):
        self[old.index] = new
        new.parent = self
        new._size = old._size  # pylint: disable=protected-access

    def flip_with(self, node, reverse=False):
        """Join with node in a new, orthogonal container."""
        container = Node()
        self.parent.replace_child(self, container)
        self.reset_size()
        for child in [node, self] if reverse else [self, node]:
            container.add_child(child)

    def add_node(self, node, mode=None):
        """Add node according to the mode.

        This can result in adding it as a child, joining with it in a new
        flipped sub-container, or splitting the space with it.
        """
        if self.is_root:
            self.add_child(node)
        elif mode is None:
            self.parent.add_child_after(node, self)
        elif mode.orient is self.parent.orient:
            if mode & AddMode.SPLIT:
                node._size = 0  # pylint: disable=protected-access
                self.parent.add_child_after(node, self)
                self._size = node._size = self.size / 2
            else:
                self.parent.add_child_after(node, self)
        else:
            self.flip_with(node)

    def _clean_restorables(self, client):
        with contextlib.suppress(KeyError):
            del self.root.restorables[client]

    def restore(self, node):
        """Restore node.

        Try to add the node in a place where a node with the same payload
        has previously been.
        """
        restorables = self.root.restorables
        try:
            parent, idx, sizes, fixed, flip = restorables[node.payload]
        except KeyError:
            raise NotRestorableError()  # pylint: disable=raise-missing-from
        if parent not in self.root:
            # Don't try to restore if parent is not part of the tree anymore
            raise NotRestorableError()
        node.reset_size()
        if flip:
            old_parent_size = parent.size
            parent.flip_with(node, reverse=(idx == 0))
            node.size, parent.size = sizes
            Node.fit_into(parent, old_parent_size)
        else:
            parent.add_child(node, idx=idx)
            node.size = sizes[0]
            if len(sizes) == 2:
                if node.siblings:
                    node.siblings[0].size = sizes[1]
                else:
                    node.reset_size()
        if not fixed:
            node.reset_size()
        del restorables[node.payload]

    def _save_restore_state(self):
        parent = self.parent
        sizes = (self.size,)
        flip = False
        if len(self.siblings) == 1:
            # If there is only one node left in the container, we need to save
            # its size too because the size will be lost.
            sizes += (self.siblings[0]._size,)  # pylint: disable=W0212
            if not self.parent.is_root:
                flip = True
                parent = self.siblings[0]
        self.root.restorables[self.payload] = (parent, self.index, sizes, self.fixed, flip)

    def move(self, direction):
        """Move this node in `direction`. Return whether node was moved."""
        if self.is_root:
            return False
        if direction.orient is self.parent.orient:
            old_idx = self.index
            new_idx = old_idx + direction.offset
            if 0 <= new_idx < len(self.parent):
                p = self.parent
                p[old_idx], p[new_idx] = p[new_idx], p[old_idx]
                return True
            new_sibling = self.parent.parent
        else:
            new_sibling = self.parent
        try:
            new_parent = new_sibling.parent
            idx = new_sibling.index
        except AttributeError:
            return False
        self.reset_size()
        self.parent.remove_child(self)
        new_parent.add_child(self, idx + (1 if direction.offset == 1 else 0))
        return True

    def move_up(self):
        return self.move(Direction.UP)

    def move_down(self):
        return self.move(Direction.DOWN)

    def move_right(self):
        return self.move(Direction.RIGHT)

    def move_left(self):
        return self.move(Direction.LEFT)

    def _move_and_integrate(self, direction):
        old_parent = self.parent
        self.move(direction)
        if self.parent is not old_parent:
            self.integrate(direction)

    def integrate(self, direction):
        if direction.orient != self.parent.orient:
            self._move_and_integrate(direction)
            return
        target_idx = self.index + direction.offset
        if target_idx < 0 or target_idx >= len(self.parent):
            self._move_and_integrate(direction)
            return
        self.reset_size()
        target = self.parent[target_idx]
        self.parent.remove_child(self)
        if target.is_leaf:
            target.flip_with(self)
        else:
            target.add_child(self)

    def integrate_up(self):
        self.integrate(Direction.UP)

    def integrate_down(self):
        self.integrate(Direction.DOWN)

    def integrate_left(self):
        self.integrate(Direction.LEFT)

    def integrate_right(self):
        self.integrate(Direction.RIGHT)

    def find_payload(self, payload):
        if self.payload is payload:
            return self
        for child in self:
            needle = child.find_payload(payload)
            if needle is not None:
                return needle
        return None


class Plasma(Layout):
    """
    A flexible tree-based layout.

    Each tree node represents a container whose children are aligned either
    horizontally or vertically. Each window is attached to a leaf of the tree
    and takes either a calculated relative amount or a custom absolute amount
    of space in its parent container. Windows can be resized, rearranged and
    integrated into other containers.

    Windows in a container will all open in the same direction. Calling
    ``lazy.layout.mode_vertical/horizontal()`` will insert a new container allowing
    windows to be added in the new direction.

    You can use the ``Plasma`` widget to show which mode will apply when opening a new
    window based on the currently focused node.

    Windows can be focused selectively by using ``lazy.layout.up/down/left/right()`` to focus
    the nearest window in that direction relative to the currently focused window.

    "Integrating" windows is best explained with an illustation. Starting with three
    windows, a, b, c. b is currently focused. Calling ``lazy.layout.integrate_left()``
    will have the following effect:

    ::

        ----------------------         ----------------------
        | a    | b    | c    |         | a        | c       |
        |      |      |      |         |          |         |
        |      |      |      |  -->    |          |         |
        |      |      |      |         |----------|         |
        |      |      |      |         | b        |         |
        |      |      |      |         |          |         |
        |      |      |      |         |          |         |
        ----------------------         ----------------------

    Finally, windows can me moved around the layout with ``lazy.layout.move_up/down/left/right()``.

    Example keybindings:

    .. code:: python

        from libqtile.config import EzKey
        from libqtile.lazy import lazy
        ...
        keymap = {
            'M-h': lazy.layout.left(),
            'M-j': lazy.layout.down(),
            'M-k': lazy.layout.up(),
            'M-l': lazy.layout.right(),
            'M-S-h': lazy.layout.move_left(),
            'M-S-j': lazy.layout.move_down(),
            'M-S-k': lazy.layout.move_up(),
            'M-S-l': lazy.layout.move_right(),
            'M-A-h': lazy.layout.integrate_left(),
            'M-A-j': lazy.layout.integrate_down(),
            'M-A-k': lazy.layout.integrate_up(),
            'M-A-l': lazy.layout.integrate_right(),
            'M-d': lazy.layout.mode_horizontal(),
            'M-v': lazy.layout.mode_vertical(),
            'M-S-d': lazy.layout.mode_horizontal_split(),
            'M-S-v': lazy.layout.mode_vertical_split(),
            'M-a': lazy.layout.grow_width(30),
            'M-x': lazy.layout.grow_width(-30),
            'M-S-a': lazy.layout.grow_height(30),
            'M-S-x': lazy.layout.grow_height(-30),
            'M-C-5': lazy.layout.size(500),
            'M-C-8': lazy.layout.size(800),
            'M-n': lazy.layout.reset_size(),
        }
        keys = [EzKey(k, v) for k, v in keymap.items()]

    Acknowledgements:
    This layout was developed by numirias and published at
    https://github.com/numirias/qtile-plasma A few minor amendments have been made
    to that layout as part of incorporating this into the main qtile codebase but the
    majority of the work is theirs.

    """

    defaults = [
        ("name", "Plasma", "Layout name"),
        ("border_normal", "#333333", "Unfocused window border color"),
        ("border_focus", "#00e891", "Focused window border color"),
        ("border_normal_fixed", "#333333", "Unfocused fixed-size window border color"),
        ("border_focus_fixed", "#00e8dc", "Focused fixed-size window border color"),
        ("border_width", 1, "Border width"),
        ("border_width_single", 0, "Border width for single window"),
        ("margin", 0, "Layout margin"),
        (
            "fair",
            False,
            "When ``False`` effort will be made to preserve nodes with a fixed size. "
            "Set to ``True`` to enable new windows to take more space from fixed size nodes.",
        ),
    ]
    # If windows are added before configure() was called, the screen size is
    # still unknown, so we need to set some arbitrary initial root dimensions
    default_dimensions = (0, 0, 1000, 1000)

    def __init__(self, **config):
        Layout.__init__(self, **config)
        self.add_defaults(Plasma.defaults)
        self.root = Node(None, *self.default_dimensions)
        self._focused = None
        self._add_mode = None
        Node.priority = Priority.BALANCED if self.fair else Priority.FIXED

    def swap(self, c1: Window, c2: Window) -> None:
        node_c1 = node_c2 = None
        for leaf in self.root.all_leafs:
            if leaf.payload is not None:
                if c1 == leaf.payload:
                    node_c1 = leaf
                elif c2 == leaf.payload:
                    node_c2 = leaf
            if node_c1 is not None and node_c2 is not None:
                node_c1.payload, node_c2.payload = node_c2.payload, node_c1.payload
                self.group.layout_all()
                self.group.focus(c1)
                return

    @staticmethod
    def convert_names(tree):
        return [Plasma.convert_names(n) if isinstance(n, list) else n.payload.name for n in tree]

    @property
    def add_mode(self):
        if self._add_mode is None:
            node = self.root_or_focused_node
            if node.width >= node.height:
                return AddMode.HORIZONTAL
            else:
                return AddMode.VERTICAL

        return self._add_mode

    @add_mode.setter
    def add_mode(self, value):
        self._add_mode = value
        # We trigger a redraw so that the different borders can be drawn based on the add_mode
        # We check self._group to avoid raising a runtime error from libqtile.layout.base
        if self._group is not None:
            hook.fire("plasma_add_mode", self)
            self.group.layout_all()

    @property
    def focused(self):
        return self._focused

    @focused.setter
    def focused(self, value):
        self._focused = value
        hook.fire("plasma_add_mode", self)

    @property
    def focused_node(self):
        return self.root.find_payload(self.focused)

    @property
    def root_or_focused_node(self):
        return self.root if self.focused_node is None else self.focused_node

    @property
    def horizontal(self):
        if self.focused_node is None:
            return True

        if self.add_mode is not None:
            if self.add_mode & AddMode.HORIZONTAL:
                return True
            else:
                return False

        if self.focused_node.parent is None:
            if self.focused_node.orient is Orient.HORIZONTAL:
                return True
            else:
                return False

        return self.focused_node.parent.horizontal

    @property
    def vertical(self):
        return not self.horizontal

    @property
    def split(self):
        if self.add_mode is not None and self.add_mode & AddMode.SPLIT:
            return True

        return False

    @expose_command
    def info(self):
        info = super().info()
        tree = self.convert_names(self.root.tree)
        info["tree"] = tree
        info["clients"] = flatten(tree)

        return info

    def clone(self, group):
        clone = copy.copy(self)
        clone._group = group
        clone.root = Node(None, *self.default_dimensions)
        clone.focused = None
        clone.add_mode = None
        return clone

    def get_windows(self):
        clients = []
        for leaf in self.root.all_leafs:
            if leaf.payload is not None:
                clients.append(leaf.payload)
        return clients

    def add_client(self, client):
        new = Node(client)
        try:
            self.root.restore(new)
        except NotRestorableError:
            self.root_or_focused_node.add_node(new, self.add_mode)
        self.add_mode = None

    def remove(self, client):
        payload = self.root.find_payload(client)
        if payload is not None:
            payload.remove()

    def configure(self, client, screen_rect):
        self.root.x = screen_rect.x
        self.root.y = screen_rect.y
        self.root.width = screen_rect.width
        self.root.height = screen_rect.height
        node = self.root.find_payload(client)
        border_width = self.border_width_single if self.root.tree == [node] else self.border_width
        border_color = getattr(
            self,
            "border_"
            + ("focus" if client.has_focus else "normal")
            + ("" if node.flexible else "_fixed"),
        )
        x, y, width, height = node.pixel_perfect
        client.place(
            x,
            y,
            width - 2 * border_width,
            height - 2 * border_width,
            border_width,
            border_color,
            margin=self.margin,
        )
        # Always keep tiles below floating windows
        client.unhide()

    def focus(self, client):
        self.focused = client
        self.root.find_payload(client).access()

    def focus_first(self):
        return self.root.first_leaf.payload

    def focus_last(self):
        return self.root.last_leaf.payload

    def focus_next(self, win):
        next_leaf = self.root.find_payload(win).next_leaf
        return None if next_leaf is self.root.first_leaf else next_leaf.payload

    def focus_previous(self, win):
        prev_leaf = self.root.find_payload(win).prev_leaf
        return None if prev_leaf is self.root.last_leaf else prev_leaf.payload

    def focus_node(self, node):
        if node is None:
            return
        self.group.focus(node.payload)

    def refocus(self):
        self.group.focus(self.focused)

    @expose_command
    def next(self):
        """Focus next window."""
        self.focus_node(self.focused_node.next_leaf)

    @expose_command
    def previous(self):
        """Focus previous window."""
        self.focus_node(self.focused_node.prev_leaf)

    @expose_command
    def recent(self):
        """Focus most recently focused window.

        (Toggles between the two latest active windows.)
        """
        nodes = [n for n in self.root.all_leafs if n is not self.focused_node]
        most_recent = max(nodes, key=lambda n: n.last_accessed)
        self.focus_node(most_recent)

    @expose_command
    def left(self):
        """Focus window to the left."""
        self.focus_node(self.focused_node.close_left)

    @expose_command
    def right(self):
        """Focus window to the right."""
        self.focus_node(self.focused_node.close_right)

    @expose_command
    def up(self):
        """Focus window above."""
        self.focus_node(self.focused_node.close_up)

    @expose_command
    def down(self):
        """Focus window below."""
        self.focus_node(self.focused_node.close_down)

    @expose_command
    def move_left(self):
        """Move current window left."""
        self.focused_node.move_left()
        self.refocus()

    @expose_command
    def move_right(self):
        """Move current window right."""
        self.focused_node.move_right()
        self.refocus()

    @expose_command
    def move_up(self):
        """Move current window up."""
        self.focused_node.move_up()
        self.refocus()

    @expose_command
    def move_down(self):
        """Move current window down."""
        self.focused_node.move_down()
        self.refocus()

    @expose_command
    def integrate_left(self):
        """Integrate current window left."""
        self.focused_node.integrate_left()
        self.refocus()

    @expose_command
    def integrate_right(self):
        """Integrate current window right."""
        self.focused_node.integrate_right()
        self.refocus()

    @expose_command
    def integrate_up(self):
        """Integrate current window up."""
        self.focused_node.integrate_up()
        self.refocus()

    @expose_command
    def integrate_down(self):
        """Integrate current window down."""
        self.focused_node.integrate_down()
        self.refocus()

    @expose_command
    def mode_horizontal(self):
        """Next window will be added horizontally."""
        self.add_mode = AddMode.HORIZONTAL

    @expose_command
    def mode_vertical(self):
        """Next window will be added vertically."""
        self.add_mode = AddMode.VERTICAL

    @expose_command
    def mode_horizontal_split(self):
        """Next window will be added horizontally, splitting space of current
        window.
        """
        self.add_mode = AddMode.HORIZONTAL | AddMode.SPLIT

    @expose_command
    def mode_vertical_split(self):
        """Next window will be added vertically, splitting space of current
        window.
        """
        self.add_mode = AddMode.VERTICAL | AddMode.SPLIT

    @expose_command
    def set_size(self, x: int):
        """Change size of current window.

        (It's recommended to use `width()`/`height()` instead.)
        """
        self.focused_node.size = x
        self.refocus()

    @expose_command
    def set_width(self, x: int):
        """Set width of current window."""
        self.focused_node.width = x
        self.refocus()

    @expose_command
    def set_height(self, x: int):
        """Set height of current window."""
        self.focused_node.height = x
        self.refocus()

    @expose_command
    def reset_size(self):
        """Reset size of current window to automatic (relative) sizing."""
        self.focused_node.reset_size()
        self.refocus()

    @expose_command
    def grow(self, x: int):
        """Grow size of current window.

        (It's recommended to use `grow_width()`/`grow_height()` instead.)
        """
        self.focused_node.size += x
        self.refocus()

    @expose_command
    def grow_width(self, x: int):
        """Grow width of current window."""
        self.focused_node.width += x
        self.refocus()

    @expose_command
    def grow_height(self, x: int):
        """Grow height of current window."""
        self.focused_node.height += x
        self.refocus()



from libqtile.backend.base.idle_notify import IdleNotifier as BaseIdleNotifier

try:
    from libqtile.backend.wayland._ffi import lib
except ModuleNotFoundError:
    from libqtile.backend.wayland.ffi_stub import lib


class IdleNotifier(BaseIdleNotifier):
    def clear_timers(self) -> None:
        for timeout in self.timeouts:
            lib.qw_server_remove_idle_timer(self.core.qw, timeout)
        super().clear_timers()

    def run(self) -> None:
        for timeout in self.timeouts:
            lib.qw_server_add_idle_timer(self.core.qw, timeout)

    def handle_timeout(self, timeout: int) -> None:
        self.fire_action(timeout)

    def handle_resume(self) -> None:
        self.fire_resume()

    def reset(self) -> None:
        self.clear_timers()
        self.start()



import json
import os
import sys
from tempfile import TemporaryDirectory

from jupyter_client.kernelspec import install_kernel_spec

kernel_json = {
    "argv": [
        sys.executable,
        "-m",
        "libqtile.interactive.iqshell_kernel",
        "-f",
        "{connection_file}",
    ],
    "display_name": "iqshell",
    "language": "qshell",
}


def main(argv=None):
    if argv is None:
        argv = []
    user = "--user" in argv or os.geteuid() != 0

    with TemporaryDirectory() as td:
        # IPython tempdir starts off as 700, not user readable
        os.chmod(td, 0o755)
        with open(os.path.join(td, "kernel.json"), "w") as f:
            json.dump(kernel_json, f, sort_keys=True)

        print("Installing IPython kernel spec")
        install_kernel_spec(td, "qshell", user=user, replace=True)


if __name__ == "__main__":
    main(sys.argv)



from libqtile.widget import base


class She(base.InLoopPollText):
    """Widget to display the Super Hybrid Engine status

    Can display either the mode or CPU speed on eeepc computers.
    """

    defaults = [
        ("device", "/sys/devices/platform/eeepc/cpufv", "sys path to cpufv"),
        ("format", "speed", 'Type of info to display "speed" or "name"'),
        ("update_interval", 0.5, "Update Time in seconds."),
    ]

    def __init__(self, **config):
        base.InLoopPollText.__init__(self, **config)
        self.add_defaults(She.defaults)
        self.modes = {
            "0x300": {"name": "Performance", "speed": "1.6GHz"},
            "0x301": {"name": "Normal", "speed": "1.2GHz"},
            "0x302": {"name": "PoswerSave", "speed": "800MHz"},
        }

    def poll(self):
        with open(self.device) as f:
            mode = f.read().strip()
        if mode in self.modes:
            return self.modes[mode][self.format]
        else:
            return mode



from __future__ import annotations

import typing

import libqtile.backend.base.window as base
from libqtile import hook, utils
from libqtile.backend.base import FloatStates
from libqtile.backend.base.window import WindowType
from libqtile.backend.wayland.drawer import Drawer
from libqtile.command.base import CommandError, expose_command
from libqtile.core.manager import Qtile
from libqtile.group import _Group
from libqtile.log_utils import logger
from libqtile.utils import ColorsType, rgb

try:
    # Continue if ffi not built, so that docs can be built without wayland deps.
    from libqtile.backend.wayland._ffi import ffi, lib
except ModuleNotFoundError:
    print("Warning: Wayland backend not built. Backend will not run.")
    # Continue if ffi not built, so that docs can be built without wayland deps.
    # Provide a stub for FFI to keep going
    from libqtile.backend.wayland.ffi_stub import ffi, lib

if typing.TYPE_CHECKING:
    from libqtile.backend.wayland.core import Core
    from libqtile.command.base import CommandObject, ItemT


class Base(base._Window):
    def __init__(self, qtile: Qtile, ptr: ffi.CData, wid: int):
        base._Window.__init__(self)
        self.qtile = qtile
        self._group: _Group | None = None
        # TODO: why are there 2 groups here?
        self._ptr = ptr
        self._wid = wid
        self._wm_class: str | None = None
        self._wm_instance: str | None = None
        self._wm_role: str | None = None
        self.bordercolor: ColorsType | None = None
        self.borderwidth = 0
        # TODO: what is this?
        self.defunct = False
        self.group: _Group | None = None
        self.core: Core = typing.cast("Core", qtile.core)

    def reparent(self, layer: int) -> None:
        if self.layer == layer:
            return
        lib.qw_view_reparent(self._ptr, layer)

    @property
    def layer(self) -> int:
        return self._ptr.layer

    @expose_command()
    def keep_above(self, enable: bool | None = None) -> None:
        is_enabled = self.layer == lib.LAYER_KEEPABOVE
        if enable is None:
            enable = not is_enabled

        if enable:
            self.reparent(lib.LAYER_KEEPABOVE)
        else:
            self.reparent(lib.LAYER_LAYOUT)

    @expose_command()
    def keep_below(self, enable: bool | None = None) -> None:
        is_enabled = self.layer == lib.LAYER_KEEPBELOW
        if enable is None:
            enable = not is_enabled
        if enable:
            self.reparent(lib.LAYER_KEEPBELOW)
            self.move_to_bottom()
        else:
            self.reparent(lib.LAYER_LAYOUT)
            self.move_to_bottom()

    @expose_command()
    def move_to_top(self) -> None:
        lib.qw_view_raise_to_top(self._ptr)

    @expose_command()
    def move_up(self, force: bool = False) -> None:
        lib.qw_view_move_up(self._ptr)

    @expose_command()
    def move_down(self, force: bool = False) -> None:
        lib.qw_view_move_down(self._ptr)

    @expose_command()
    def move_to_bottom(self) -> None:
        lib.qw_view_lower_to_bottom(self._ptr)

    @expose_command()
    def bring_to_front(self) -> None:
        self.reparent(lib.LAYER_BRINGTOFRONT)
        lib.qw_view_raise_to_top(self._ptr)

    @property
    def wid(self) -> int:
        return self._wid

    @property
    def x(self) -> int:
        return self._ptr.x

    @x.setter
    def x(self, x: int) -> None:
        self._ptr.x = x

    @property
    def y(self) -> int:
        return self._ptr.y

    @y.setter
    def y(self, y: int) -> None:
        self._ptr.y = y

    @property
    def width(self) -> int:
        return self._ptr.width

    @width.setter
    def width(self, width: int) -> None:
        self._ptr.width = width

    @property
    def height(self) -> int:
        return self._ptr.height

    @height.setter
    def height(self, height: int) -> None:
        self._ptr.height = height

    @property
    def urgent(self) -> bool:
        return self._ptr.urgent

    @urgent.setter
    def urgent(self, urgent: bool) -> None:
        self._ptr.urgent = urgent

    def kill(self) -> None:
        self._ptr.kill(self._ptr)

    def hide(self) -> None:
        self._ptr.hide(self._ptr)
        self.core.check_inhibited()

    def unhide(self) -> None:
        self._ptr.unhide(self._ptr)
        self.core.check_inhibited()

    @expose_command()
    def place(
        self,
        x: int | None,
        y: int | None,
        width: int | None,
        height: int | None,
        borderwidth: int | None = None,
        bordercolor: ColorsType | None = None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        # Adjust the placement to account for layout margins, if there are any.
        # TODO: is respect_hints only for X11?
        assert ffi is not None
        if x is None:
            x = self.x
        if y is None:
            y = self.y
        if bordercolor is None:
            bordercolor = self.bordercolor
        if borderwidth is None:
            borderwidth = borderwidth or 0
        if width is None:
            width = self.width
        if height is None:
            height = self.height
        if margin is not None:
            if isinstance(margin, int):
                margin = [margin] * 4
            x += margin[3]
            y += margin[0]
            width -= margin[1] + margin[3]
            height -= margin[0] + margin[2]

        # TODO: respect hints

        if self.group is not None and self.group.screen is not None:
            self.float_x = x - self.group.screen.x
            self.float_y = y - self.group.screen.y
        n = 0
        c_layers = ffi.NULL
        if bordercolor is not None:
            if isinstance(bordercolor, list):
                # multiple border colors
                if all(isinstance(x, str) or isinstance(x, tuple) for x in bordercolor):
                    colors = bordercolor
                else:
                    pass
                    # TODO: Check validation logic, handle this case
            else:
                colors = [bordercolor]
            n = len(colors)

            # Allocate array of qw_border
            c_layers = ffi.new(f"struct qw_border[{n}]")

            base_width = borderwidth // n
            remainder = borderwidth % n

            for i, bc in enumerate(colors):
                # Set type RECT
                c_layers[i].type = lib.QW_BORDER_RECT

                # Set width, distribute remainder
                c_layers[i].width = base_width + (1 if i < remainder else 0)

                # Convert python color to RGBA float[4]
                rgba = rgb(bc)

                # Copy RGBA into c struct color field
                # Each edge is set to the same colour
                for side in range(4):
                    for j in range(4):
                        c_layers[i].rect.color[side][j] = rgba[j]

        self.bordercolor = bordercolor
        self.borderwidth = borderwidth
        self._ptr.place(self._ptr, x, y, width, height, c_layers, n, int(above))

    @expose_command()
    def focus(self, warp: bool = True) -> None:
        self.core.focus_window(self)

        if warp and self.qtile.config.cursor_warp:
            self.qtile.core.warp_pointer(
                self.x + self.width // 2,
                self.y + self.height // 2,
            )

        if self.urgent:
            self.urgent = False

        if self.group and self.group.current_window is not self:
            self.group.focus(self)

        hook.fire("client_focus", self)


class Internal(Base, base.Internal):
    def __init__(self, qtile: Qtile, ptr: ffi.CData, wid: int):
        Base.__init__(self, qtile, lib.qw_internal_view_get_base(ptr), wid)
        base.Internal.__init__(self)
        ptr.base.wid = wid
        self._internal_ptr = ptr
        self._killed = False

    @property
    def surface(self) -> ffi.CData:
        return ffi.cast("void *", self._internal_ptr.image_surface)

    @property
    def scale(self) -> float:
        return self._internal_ptr.scale

    def finalize(self) -> None:
        self.hide()

    def create_drawer(self, width: int, height: int) -> Drawer:
        """Create a Drawer that draws to this window."""
        return Drawer(self, width, height)

    def set_buffer_with_damage(self, offsetx: int, offsety: int, width: int, height: int) -> None:
        lib.qw_internal_view_set_buffer_with_damage(
            self._internal_ptr, offsetx, offsety, width, height
        )

    @expose_command()
    def kill(self) -> None:
        if not self._killed:
            self._ptr.kill(self._ptr)
            self._killed = True
        if self.wid in self.qtile.windows_map:
            # It will be present during config reloads; absent during shutdown as this
            # will follow graceful_shutdown
            del self.qtile.windows_map[self.wid]

    @expose_command()
    def info(self) -> dict:
        """Return a dictionary of info."""
        return dict(
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            id=self.wid,
        )


@ffi.def_extern()
def request_focus_cb(userdata: ffi.CData) -> int:
    win = ffi.from_handle(userdata)
    if win.handle_request_focus():
        return 1
    return 0


@ffi.def_extern()
def request_close_cb(userdata: ffi.CData) -> int:
    win = ffi.from_handle(userdata)
    if win.handle_request_close():
        return 1
    return 0


@ffi.def_extern()
def request_fullscreen_cb(fullscreen: bool, userdata: ffi.CData) -> int:
    win = ffi.from_handle(userdata)
    if win.handle_request_fullscreen(fullscreen):
        return 1
    return 0


@ffi.def_extern()
def request_maximize_cb(maximize: bool, userdata: ffi.CData) -> int:
    win = ffi.from_handle(userdata)
    if win.handle_request_maximize(maximize):
        return 1
    return 0


@ffi.def_extern()
def request_minimize_cb(minimize: bool, userdata: ffi.CData) -> int:
    win = ffi.from_handle(userdata)
    if win.handle_request_minimize(minimize):
        return 1
    return 0


@ffi.def_extern()
def set_title_cb(title: ffi.CData, userdata: ffi.CData) -> None:
    win = ffi.from_handle(userdata)
    win.handle_set_title(ffi.string(title).decode())


@ffi.def_extern()
def set_app_id_cb(app_id: ffi.CData, userdata: ffi.CData) -> None:
    win = ffi.from_handle(userdata)
    win.handle_set_app_id(ffi.string(app_id).decode())


class Window(Base, base.Window):
    def __init__(self, qtile: Qtile, ptr: ffi.CData, wid: int):
        Base.__init__(self, qtile, ptr, wid)
        base.Window.__init__(self)

        self.float_x: int | None = None
        self.float_y: int | None = None
        self._float_width: int = 0
        self._float_height: int = 0
        self._float_state = FloatStates.NOT_FLOATING
        # TODO: destroy?
        self._userdata = ffi.new_handle(self)
        ptr.cb_data = self._userdata
        ptr.request_focus_cb = lib.request_focus_cb
        ptr.request_close_cb = lib.request_close_cb
        ptr.request_maximize_cb = lib.request_maximize_cb
        ptr.request_minimize_cb = lib.request_minimize_cb
        ptr.request_fullscreen_cb = lib.request_fullscreen_cb
        ptr.set_title_cb = lib.set_title_cb
        ptr.set_app_id_cb = lib.set_app_id_cb

    def handle_request_focus(self) -> bool:
        logger.debug("Focusing window from external request")
        return self.activate()

    def handle_request_close(self) -> bool:
        self.kill()
        return True

    def handle_request_fullscreen(self, fullscreen: bool) -> bool:
        if self.qtile.config.auto_fullscreen:
            if self.fullscreen != fullscreen:
                self.fullscreen = fullscreen
                return True
        return False

    def handle_request_maximize(self, maximize: bool) -> bool:
        self.maximized = maximize
        return True

    def handle_request_minimize(self, minimize: bool) -> bool:
        self.minimized = minimize
        return True

    def handle_set_title(self, title: str) -> None:
        logger.debug("Signal: xdgwindow set_title")
        if title != self.name:
            self.name = title
            # TODO: Handle foreign-toplevel-management?
            hook.fire("client_name_updated", self)

    def handle_set_app_id(self, app_id: str) -> None:
        logger.debug("Signal: xdgwindow set_app_id")
        self._wm_class = app_id
        # TODO: Handle foreign-toplevel-management?

    def get_wm_class(self) -> list | None:
        wm_class = []
        if self._wm_instance:
            wm_class.append(self._wm_instance)
        if self._wm_class:
            wm_class.append(self._wm_class)

        return wm_class or None

    def get_wm_role(self) -> str | None:
        return self._wm_role or None

    @expose_command()
    def is_visible(self) -> bool:
        return lib.qw_view_is_visible(self._ptr)

    @expose_command()
    def static(
        self,
        screen: int | None = None,
        x: int | None = None,
        y: int | None = None,
        width: int | None = None,
        height: int | None = None,
    ) -> None:
        # The concrete Window class must fire the client_managed hook after it's
        # completed any custom logic.
        self.defunct = True
        if self.group:
            self.group.remove(self)

        # Keep track of user-specified geometry to support X11.
        # Respect configure requests only if these are `None` here.
        conf_x = x
        conf_y = y
        conf_width = width
        conf_height = height

        if x is None:
            x = self.x + self.borderwidth
        if y is None:
            y = self.y + self.borderwidth
        if width is None:
            width = self.width
        if height is None:
            height = self.height

        win = self._to_static(conf_x, conf_y, conf_width, conf_height)

        # Remove references to original window
        self._userdata = None

        # TODO: pass over ftm

        if screen is not None:
            win.screen = self.qtile.screens[screen]
        win.unhide()
        win.place(x, y, width, height, 0, None)
        self.qtile.windows_map[self.wid] = win

        # TODO: pointer constraints

        hook.fire("client_managed", win)

    def _to_static(
        self, x: int | None, y: int | None, width: int | None, height: int | None
    ) -> Static:
        return Static(
            self.qtile,
            self._ptr,
            self._wid,
        )

    @expose_command()
    def togroup(
        self, group_name: str | None = None, switch_group: bool = False, toggle: bool = False
    ) -> None:
        """
        Move window to a specified group

        Also switch to that group if switch_group is True.

        If `toggle` is True and and the specified group is already on the screen,
        use the last used group as target instead.
        """
        if group_name is None:
            group = self.qtile.current_group
        else:
            if group_name not in self.qtile.groups_map:
                raise CommandError(f"No such group: {group_name}")
            group = self.qtile.groups_map[group_name]

        if self.group is group:
            if toggle and self.group.screen.previous_group:
                group = self.group.screen.previous_group
            else:
                return

        self.hide()
        if self.group:
            if self.group.screen:
                # for floats remove window offset
                self.x -= self.group.screen.x
            group_ref = self.group
            self.group.remove(self)
            # delete groups with `persist=False`
            if (
                not self.qtile.dgroups.groups_map[group_ref.name].persist
                and len(group_ref.windows) <= 1
            ):
                # set back original group so _del() can grab it
                self.group = group_ref
                self.qtile.dgroups._del(self)
                self.group = None

        if group.screen and self.x < group.screen.x:
            self.x += group.screen.x
        group.add(self)
        if switch_group:
            group.toscreen(toggle=toggle)

    def _items(self, name: str) -> ItemT:
        if name == "group":
            return True, []
        if name == "layout":
            if self.group:
                return True, list(range(len(self.group.layouts)))
            return None
        if name == "screen":
            if self.group and self.group.screen:
                return True, []
        return None

    def _select(self, name: str, sel: str | int | None) -> CommandObject | None:
        if name == "group":
            return self.group
        elif name == "layout":
            if sel is None:
                return self.group.layout if self.group else None
            else:
                return utils.lget(self.group.layouts, int(sel)) if self.group else None
        elif name == "screen":
            return self.group.screen if self.group else None
        return None

    @property
    def group(self) -> _Group | None:
        return self._group

    @group.setter
    def group(self, group: _Group | None) -> None:
        self._group = group

    @property
    def skip_taskbar(self) -> bool:
        return bool(self._ptr.skip_taskbar)

    @skip_taskbar.setter
    def skip_taskbar(self, value: bool) -> None:
        self._ptr.skip_taskbar = value

    @expose_command()
    def get_position(self) -> tuple[int, int]:
        return int(self._ptr.x), int(self._ptr.y)

    @expose_command()
    def get_size(self) -> tuple[int, int]:
        return int(self._ptr.width), int(self._ptr.height)

    def has_fixed_size(self) -> bool:
        return self._ptr.has_fixed_size(self._ptr)

    def get_pid(self) -> int:
        return int(self._ptr.get_pid(self._ptr))

    def get_wm_type(self) -> str:
        return ffi.string(self._ptr.get_wm_type(self._ptr)).decode()

    def is_transient_for(self) -> WindowType | None:
        wid = int(self._ptr.get_parent(self._ptr))
        return self.qtile.windows_map.get(wid, None)

    def get_new_layer(self, state: FloatStates) -> int:
        if self.qtile.config.floats_kept_above and state == FloatStates.FLOATING:
            return lib.LAYER_KEEPABOVE
        if state == FloatStates.MAXIMIZED:
            return lib.LAYER_MAX
        if state == FloatStates.FULLSCREEN:
            return lib.LAYER_FULLSCREEN
        return lib.LAYER_LAYOUT

    @expose_command()
    def move_up(self, force: bool = False) -> None:
        if force and self.layer == lib.LAYER_KEEPBELOW:
            new_layer = self.get_new_layer(self._float_state)
            self.reparent(new_layer)
        lib.qw_view_move_up(self._ptr)

    @expose_command()
    def move_down(self, force: bool = False) -> None:
        if force and self.layer == lib.LAYER_KEEPAOVE:
            new_layer = self.get_new_layer(self._float_state)
            self.reparent(new_layer)
        lib.qw_view_move_down(self._ptr)

    @expose_command()
    def info(self) -> dict:
        """Return a dictionary of info."""
        # TODO: complete implementation
        float_info = {
            "x": self.float_x,
            "y": self.float_y,
            "width": self._float_width,
            "height": self._float_height,
        }
        return dict(
            name=self.name,
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            group=self.group.name if self.group else None,
            id=self.wid,
            wm_class=self.get_wm_class(),
            # shell can be either "XDG" or "XWayland" or "layer"?
            shell=ffi.string(self._ptr.shell).decode() if self._ptr.shell != ffi.NULL else "",
            float_info=float_info,
            floating=self._float_state != FloatStates.NOT_FLOATING,
            maximized=self._float_state == FloatStates.MAXIMIZED,
            minimized=self._float_state == FloatStates.MINIMIZED,
            fullscreen=self._float_state == FloatStates.FULLSCREEN,
        )

    @property
    def floating(self) -> bool:
        return self._float_state != FloatStates.NOT_FLOATING

    @floating.setter
    def floating(self, do_float: bool) -> None:
        if do_float and self._float_state == FloatStates.NOT_FLOATING:
            if self.is_placed():
                screen = self.group.screen  # type: ignore[union-attr] # see is_placed()
                if not self._float_width:  # These might start as 0
                    self._float_width = self.width
                    self._float_height = self.height
                self._reconfigure_floating(
                    screen.x + self.float_x,
                    screen.y + self.float_y,
                    self._float_width,
                    self._float_height,
                )
            else:
                # if we are setting floating early, e.g. from a hook, we don't have a screen yet
                self._float_state = FloatStates.FLOATING
            if self.layer != lib.LAYER_KEEPABOVE and self.qtile.config.floats_kept_above:
                self.keep_above(enable=True)
        elif (not do_float) and self._float_state != FloatStates.NOT_FLOATING:
            self.reparent(lib.LAYER_LAYOUT)
            self._update_fullscreen(False)
            self._update_maximized(False)
            self._update_minimized(False)
            if self._float_state == FloatStates.FLOATING:
                # store last size
                self._float_width = self.width
                self._float_height = self.height
            self._float_state = FloatStates.NOT_FLOATING
            if self.group:
                self.group.mark_floating(self, False)
            hook.fire("float_change")

    @property
    def fullscreen(self) -> bool:
        return self._float_state == FloatStates.FULLSCREEN

    @fullscreen.setter
    def fullscreen(self, do_full: bool) -> None:
        if do_full and self._float_state != FloatStates.FULLSCREEN:
            screen = (self.group and self.group.screen) or self.qtile.find_closest_screen(
                self.x, self.y
            )

            if self._float_state not in (FloatStates.MAXIMIZED, FloatStates.FULLSCREEN):
                self._save_geometry()

            bw = self.group.floating_layout.fullscreen_border_width if self.group else 0
            self._reconfigure_floating(
                screen.x,
                screen.y,
                screen.width - 2 * bw,
                screen.height - 2 * bw,
                new_float_state=FloatStates.FULLSCREEN,
            )
        elif self._float_state == FloatStates.FULLSCREEN:
            self._restore_geometry()
            self.floating = False
            self._update_fullscreen(False)

    def _update_fullscreen(self, do_full: bool) -> None:
        if do_full != (self._float_state == FloatStates.FULLSCREEN):
            self._ptr.update_fullscreen(self._ptr, do_full)

        if self.group and self.group.screen:
            screen = self.group.screen
        else:
            screen = None

        self.qtile.core.check_screen_fullscreen_background(screen)
        self.core.check_inhibited()

    @property
    def maximized(self) -> bool:
        return self._float_state == FloatStates.MAXIMIZED

    @maximized.setter
    def maximized(self, do_maximize: bool) -> None:
        if do_maximize:
            screen = (self.group and self.group.screen) or self.qtile.find_closest_screen(
                self.x, self.y
            )

            if self._float_state not in (FloatStates.MAXIMIZED, FloatStates.FULLSCREEN):
                self._save_geometry()

            bw = self.group.floating_layout.max_border_width if self.group else 0
            self._reconfigure_floating(
                screen.dx,
                screen.dy,
                screen.dwidth - 2 * bw,
                screen.dheight - 2 * bw,
                new_float_state=FloatStates.MAXIMIZED,
            )
        else:
            if self._float_state == FloatStates.MAXIMIZED:
                self._restore_geometry()
                self.floating = False

    def _update_maximized(self, do_max: bool) -> None:
        if do_max != (self._float_state == FloatStates.MAXIMIZED):
            self._ptr.update_maximized(self._ptr, do_max)

    @property
    def minimized(self) -> bool:
        return self._float_state == FloatStates.MINIMIZED

    @minimized.setter
    def minimized(self, do_minimize: bool) -> None:
        if do_minimize:
            if self._float_state != FloatStates.MINIMIZED:
                self._reconfigure_floating(new_float_state=FloatStates.MINIMIZED)
        else:
            if self._float_state == FloatStates.MINIMIZED:
                self.floating = False

    def _update_minimized(self, do_min: bool) -> None:
        if do_min != (self._float_state == FloatStates.MINIMIZED):
            self._ptr.update_minimized(self._ptr, do_min)

    def _reconfigure_floating(
        self,
        x: int | None = None,
        y: int | None = None,
        w: int | None = None,
        h: int | None = None,
        new_float_state: FloatStates = FloatStates.FLOATING,
    ) -> None:
        if self._float_state != new_float_state:
            self._float_state = new_float_state
            self.reparent(self.get_new_layer(new_float_state))
            if self.group:  # may be not, if it's called from hook
                self.group.mark_floating(self, True)
            self._update_fullscreen(new_float_state == FloatStates.FULLSCREEN)
            hook.fire("float_change")
        self._update_fullscreen(new_float_state == FloatStates.FULLSCREEN)
        self._update_maximized(new_float_state == FloatStates.MAXIMIZED)
        self._update_minimized(new_float_state == FloatStates.MINIMIZED)
        if new_float_state == FloatStates.MINIMIZED:
            self.hide()
        else:
            self.place(
                x, y, w, h, self.borderwidth, self.bordercolor, above=True, respect_hints=True
            )

    def _tweak_float(
        self,
        x: int | None = None,
        y: int | None = None,
        dx: int = 0,
        dy: int = 0,
        w: int | None = None,
        h: int | None = None,
        dw: int = 0,
        dh: int = 0,
    ) -> None:
        if x is None:
            x = self.x
        x += dx

        if y is None:
            y = self.y
        y += dy

        if w is None:
            w = self.width
        w += dw

        if h is None:
            h = self.height
        h += dh

        if h < 0:
            h = 0
        if w < 0:
            w = 0

        screen = self.qtile.find_closest_screen(x + w // 2, y + h // 2)
        if self.group and screen is not None and screen != self.group.screen:
            self.group.remove(self, force=True)
            screen.group.add(self, force=True)
            self.qtile.focus_screen(screen.index)

        self._reconfigure_floating(x, y, w, h)

    @expose_command()
    def move_floating(self, dx: int, dy: int) -> None:
        self._tweak_float(dx=dx, dy=dy)

    @expose_command()
    def resize_floating(self, dw: int, dh: int) -> None:
        self._tweak_float(dw=dw, dh=dh)

    @expose_command()
    def set_position_floating(self, x: int, y: int) -> None:
        self._tweak_float(x=x, y=y)

    @expose_command()
    def set_position(self, x: int, y: int) -> None:
        if self.floating:
            self._tweak_float(x=x, y=y)
            return

        if self.group:
            cx = self.core.qw_cursor.cursor.x
            cy = self.core.qw_cursor.cursor.y
            for window in self.group.windows:
                if (
                    window is not self
                    and not window.floating
                    and window.x <= cx <= (window.x + window.width)
                    and window.y <= cy <= (window.y + window.height)
                ):
                    self.group.layout.swap(self, window)
                    return

    @expose_command()
    def set_size_floating(self, w: int, h: int) -> None:
        self._tweak_float(w=w, h=h)

    @expose_command()
    def toggle_floating(self) -> None:
        self.floating = not self.floating

    @expose_command()
    def enable_floating(self) -> None:
        self.floating = True

    @expose_command()
    def disable_floating(self) -> None:
        self.floating = False

    @expose_command()
    def toggle_maximize(self) -> None:
        self.maximized = not self.maximized

    @expose_command()
    def toggle_minimize(self) -> None:
        self.minimized = not self.minimized

    @expose_command()
    def toggle_fullscreen(self) -> None:
        self.fullscreen = not self.fullscreen

    @expose_command()
    def enable_fullscreen(self) -> None:
        self.fullscreen = True

    @expose_command()
    def disable_fullscreen(self) -> None:
        self.fullscreen = False


class Static(Base, base.Static):
    def __init__(self, qtile: Qtile, ptr: ffi.CData, wid: int):
        Base.__init__(self, qtile, ptr, wid)
        base.Static.__init__(self)
        self.screen = qtile.current_screen
        self.x = 0
        self.y = 0
        self._width = 0
        self._height = 0
        # TODO: opacity, idle_inhibitors, ftm

        self._userdata = ffi.new_handle(self)
        ptr.cb_data = self._userdata
        ptr.request_focus_cb = ffi.NULL
        ptr.request_close_cb = ffi.NULL
        ptr.request_maximize_cb = ffi.NULL
        ptr.request_minimize_cb = ffi.NULL
        ptr.request_fullscreen_cb = ffi.NULL
        ptr.set_title_cb = lib.set_title_cb
        ptr.set_app_id_cb = lib.set_app_id_cb

        if self._ptr.title != ffi.NULL:
            self.name = ffi.string(self._ptr.title).decode()
        if self._ptr.app_id != ffi.NULL:
            self._wm_class = ffi.string(self._ptr.app_id).decode()

    def handle_set_title(self, title: str) -> None:
        logger.debug("Signal: static window set_title")
        if title != self.name:
            self.name = title
            # TODO: Handle foreign-toplevel-management?
            hook.fire("client_name_updated", self)

    def handle_set_app_id(self, app_id: str) -> None:
        logger.debug("Signal: static window set_app_id")
        self._wm_class = app_id
        # TODO: Handle foreign-toplevel-management?

    @property
    def wid(self) -> int:
        return self._wid

    def place(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
        borderwidth: int,
        bordercolor: ColorsType | None,
        above: bool = False,
        margin: int | list[int] | None = None,
        respect_hints: bool = False,
    ) -> None:
        self.x = x
        self.y = y
        self._width = width
        self._height = height

        n = 0
        self._ptr.place(self._ptr, x, y, width, height, ffi.NULL, n, int(above))

    @expose_command()
    def info(self) -> dict:
        """Return a dictionary of info."""
        info = base.Static.info(self)
        info["shell"] = (
            ffi.string(self._ptr.shell).decode() if self._ptr.shell != ffi.NULL else ""
        )
        return info



import libcst as cst
import libcst.matchers as m
from libcst import codemod
from libcst.codemod.visitors import AddImportsVisitor

from libqtile.scripts.migrations._base import (
    Check,
    MigrationTransformer,
    _QtileMigrator,
    add_migration,
)

try:
    import isort

    can_sort = True
except ImportError:
    can_sort = False


class MatchRegexTransformer(MigrationTransformer):
    def __init__(self, *args, **kwargs):
        self.needs_import = False
        MigrationTransformer.__init__(self, *args, **kwargs)

    @m.call_if_inside(
        m.Call(func=m.Name("Match")) | m.Call(func=m.Attribute(attr=m.Name("Match")))
    )
    @m.leave(m.Arg(value=m.List(), keyword=m.Name()))
    def update_match_args(self, original_node, updated_node) -> cst.Arg:
        """Changes positional  argumentto 'widgets' kwargs."""
        name = original_node.keyword.value
        if not (all(isinstance(e.value, cst.SimpleString) for e in original_node.value.elements)):
            self.lint(
                original_node,
                f"The {name} keyword uses a list but the migrate script can not fix this as "
                f"not all elements are strings.",
            )
            return updated_node

        joined_text = "|".join(e.value.value.strip("'\"") for e in original_node.value.elements)
        regex = cst.parse_expression(f"""re.compile(r"^({joined_text})$")""")
        self.lint(
            original_node,
            f"The use of a list for the {name} argument is deprecated. "
            f"Please replace with 're.compile(r\"^({joined_text})$\")'.",
        )
        self.needs_import = True
        return updated_node.with_changes(value=regex)


class MatchListRegex(_QtileMigrator):
    ID = "MatchListRegex"

    SUMMARY = "Updates Match objects using lists"

    HELP = """
    The use of lists in ``Match`` objects is deprecated and should be
    replaced with a regex.

    For example:

    .. code:: python

        Match(wm_class=["one", "two"])

    should be changed to:

    .. code::

        Match(wm_class=re.compile(r"^(one|two)$"))

    """

    AFTER_VERSION = "0.23.0"

    TESTS = [
        Check(
            """
            from libqtile.config import Match

            Match(wm_class=["one", "two"])
            """,
            """
            import re

            from libqtile.config import Match

            Match(wm_class=re.compile(r"^(one|two)$"))
            """,
        ),
    ]

    def run(self, original_node):
        # Run the base transformer first...
        # This fixes simple replacements
        transformer = MatchRegexTransformer()
        updated = original_node.visit(transformer)

        # Update details of linting
        self.update_lint(transformer)

        # Check if we replaced any lists and now need to add re import
        if transformer.needs_import:
            # Create a context to store details of changes
            context = codemod.CodemodContext()
            AddImportsVisitor.add_needed_import(context, "re")
            add_visitor = AddImportsVisitor(context)

            # Run the transformations
            updated = add_visitor.transform_module(updated)

            if can_sort:
                updated = cst.parse_module(isort.code(updated.code))
            else:
                print("Unable to sort import lines.")

        return original_node, updated


add_migration(MatchListRegex)



import re
import shlex
from subprocess import PIPE, Popen
from typing import Any

from libqtile import configurable
from libqtile.log_utils import logger

RGB = re.compile(r"^#?([a-fA-F0-9]{3}|[a-fA-F0-9]{6})$")


class _Extension(configurable.Configurable):
    """Base Extension class"""

    installed_extensions = []  # type: list

    defaults = [
        ("font", "sans", "defines the font name to be used"),
        ("fontsize", None, "defines the font size to be used"),
        ("background", None, "defines the normal background color (#RGB or #RRGGBB)"),
        ("foreground", None, "defines the normal foreground color (#RGB or #RRGGBB)"),
        ("selected_background", None, "defines the selected background color (#RGB or #RRGGBB)"),
        ("selected_foreground", None, "defines the selected foreground color (#RGB or #RRGGBB)"),
    ]

    def __init__(self, **config):
        configurable.Configurable.__init__(self, **config)
        self.add_defaults(_Extension.defaults)
        _Extension.installed_extensions.append(self)

    def _check_colors(self):
        """
        dmenu needs colours to be in #rgb or #rrggbb format.

        Checks colour value, removes invalid values and adds # if missing.

        NB This should not be called in _Extension.__init__ as _Extension.global_defaults
        may not have been set at this point.
        """
        for c in ["background", "foreground", "selected_background", "selected_foreground"]:
            col = getattr(self, c, None)
            if col is None:
                continue

            if not isinstance(col, str) or not RGB.match(col):
                logger.warning(
                    "Invalid extension '%s' color: %s. Must be #RGB or #RRGGBB string.", c, col
                )
                setattr(self, c, None)
                continue

            if not col.startswith("#"):
                col = f"#{col}"
                setattr(self, c, col)

    def _configure(self, qtile):
        self.qtile = qtile
        self._check_colors()

    def run(self):
        """
        This method must be implemented by the subclasses.
        """
        raise NotImplementedError()


class RunCommand(_Extension):
    """
    Run an arbitrary command.

    Mostly useful as a superclass for more specific extensions that need to
    interact with the qtile object.

    Also consider simply using lazy.spawn() or writing a
    `client <https://docs.qtile.org/en/latest/manual/commands/advanced.html#client-server-scripting-model>`_.
    """

    defaults: list[tuple[str, Any, str]] = [
        # NOTE: Do not use a list as a default value, since it would be shared
        #       among all the objects inheriting this class, and if one of them
        #       modified it, all the other objects would see the modified list;
        #       use a string or a tuple instead, which are immutable
        ("command", None, "the command to be launched (string or list with arguments)"),
    ]

    def __init__(self, **config):
        _Extension.__init__(self, **config)
        self.add_defaults(RunCommand.defaults)
        self.configured_command = None

    def run(self):
        """
        An extension can inherit this class, define configured_command and use
        the process object by overriding this method and using super():

        .. code-block:: python

            def _configure(self, qtile):
                Superclass._configure(self, qtile)
                self.configured_command = "foo --bar"

            def run(self):
                process = super(Subclass, self).run()
        """
        if self.configured_command:
            if isinstance(self.configured_command, str):
                self.configured_command = shlex.split(self.configured_command)
            # Else assume that self.configured_command is already a sequence
        else:
            self.configured_command = self.command
        return Popen(self.configured_command, stdout=PIPE, stdin=PIPE)



from libqtile.backend import base
from libqtile.extension.dmenu import Dmenu
from libqtile.scratchpad import ScratchPad


class WindowList(Dmenu):
    """
    Give vertical list of all open windows in dmenu. Switch to selected.
    """

    defaults = [
        ("item_format", "{group}.{id}: {window}", "the format for the menu items"),
        (
            "all_groups",
            True,
            "If True, list windows from all groups; otherwise only from the current group",
        ),
        ("dmenu_lines", "80", "Give lines vertically. Set to None get inline"),
    ]

    def __init__(self, **config):
        Dmenu.__init__(self, **config)
        self.add_defaults(WindowList.defaults)

    def list_windows(self):
        id = 0
        self.item_to_win = {}

        if self.all_groups:
            windows = [w for w in self.qtile.windows_map.values() if isinstance(w, base.Window)]
        else:
            windows = self.qtile.current_group.windows

        for win in windows:
            if win.group and not isinstance(win.group, ScratchPad):
                item = self.item_format.format(
                    group=win.group.label or win.group.name, id=id, window=win.name
                )
                self.item_to_win[item] = win
                id += 1

    def run(self):
        self.list_windows()
        out = super().run(items=self.item_to_win.keys())

        try:
            sout = out.rstrip("\n")
        except AttributeError:
            # out is not a string (for example it's a Popen object returned
            # by super(WindowList, self).run() when there are no menu items to
            # list
            return

        try:
            win = self.item_to_win[sout]
        except KeyError:
            # The selected window got closed while the menu was open?
            return

        screen = self.qtile.current_screen
        screen.set_group(win.group)
        win.group.focus(win)



import subprocess
from shutil import which

from libqtile.command.base import expose_command
from libqtile.log_utils import logger
from libqtile.widget.base import _TextBox


class GammaGroup:
    """
    GammaGroup is a helper class to group redshift gamma settings.
    """

    def __init__(self, red: float, green: float, blue: float):
        self.gamma_red = red
        self.gamma_green = green
        self.gamma_blue = blue

    def _format_gamma(self, format_txt: str) -> str:
        return format_txt.format(
            gamma_red=self.gamma_red,
            gamma_green=self.gamma_green,
            gamma_blue=self.gamma_blue,
        )

    def _redshift_fmt(self) -> str:
        return f"{self.gamma_red}:{self.gamma_green}:{self.gamma_blue}"

    def __repr__(self) -> str:
        return f"Gamma: {self._redshift_fmt()}"

    def __str__(self) -> str:
        return (
            f"GammaGroup(red={self.gamma_red}, green={self.gamma_green}, blue={self.gamma_blue})"
        )


class Redshift(_TextBox):
    """
    Redshift widget provides the following functionality:
      - Call redshift with a specific brightness, temperature and gamma config without using your location.
      - Increase/Decrease the brightness/temperature

    The redshift command can be called by just left-clicking the widget and disabling it the same way.

    If the widget is enabled, scrolling through the widget will show the settings:
    - brightness (limited to 2 dp)
    - gamma (can't be increased/decreased)
    - temperature (limited to 2 dp)
    - finally back to the default enabled/disabled text

    When at the temperature/brightness settings, the left-click/right-click mouse buttons can be used to
    increase/decrease respectively.

    Widget requirements: redshift_

    .. _redshift: https://github.com/jonls/redshift
    """

    defaults = [
        (
            "brightness",
            1.0,
            "Redshift brightness. Brightness has a lower bound of 0.1 and an upper bound of 1.0",
        ),
        ("brightness_step", 0.1, "The amount to increase/decrease the brightness by. "),
        (
            "disabled_txt",
            "󱠃",
            "Text to show when redshift is disabled. NOTE: by default, a nerd icon is used. "
            "Available fields: 'brightness' brightness to set when redshift is enabled, "
            "'is_enabled' boolean to state whether the widget is enabled or not, "
            "'gamma_blue' gamma blue value to set when redshift is enabled, "
            "'gamma_green' gamma green value to set when redshift is enabled, "
            "'gamma_red' gamma red value to set when redshift is enabled, "
            "'temperature' temperature to set when redshift is enabled, ",
        ),
        (
            "enabled_txt",
            "󰛨",
            "Text to show when redshift is disabled. NOTE: by default, a nerd icon is used. "
            "Available fields: see disabled_txt's available fields for all of them",
        ),
        (
            "gamma_red",
            1.0,
            "Redshift gamma red setting. "
            "gamma_red has a lower bound of 0.1 and an upper bound of 10.0 .",
        ),
        (
            "gamma_blue",
            1.0,
            "Redshift gamma blue. "
            "gamma_blue has a lower bound of 0.1 and an upper bound of 10.0 .",
        ),
        (
            "gamma_green",
            1.0,
            "Redshift gamma green. "
            "gamma_green has a lower bound of 0.1 and an upper bound of 10.0 .",
        ),
        ("font", "sans", "Default font"),
        ("fontsize", 20, "Font size"),
        ("foreground", "ffffff", "Font colour for information text"),
        ("redshift_path", which("redshift"), "Path to redshift executable"),
        (
            "temperature",
            1700,
            "Redshift temperature to set when enabled. "
            "Temperature has a lower bound of 1000 and an upper bound of 25000.",
        ),
        (
            "temperature_step",
            100,
            "The amount to increase/decrease the temperature by.",
        ),
        (
            "temperature_fmt",
            "Temperature: {temperature}",
            "Text to display when showing temperature text. "
            "Available fields: 'temperature' temperature to set when redshift is enabled. ",
        ),
        (
            "brightness_fmt",
            "Brightness: {brightness}",
            "Text to display when showing brightness text. "
            "Available fields: 'brightness' brightness to set when redshift is enabled. ",
        ),
        (
            "gamma_fmt",
            "Gamma: {gamma_red}:{gamma_green}:{gamma_blue}",
            "Text to display when showing gamma text. "
            "Available fields: 'gamma_red' gamma red value to set when redshift is enabled, "
            "'gamma_blue' gamma blue value to set when redshift is enabled, "
            "'gamma_green' gamma green value to set when redshift is enabled. ",
        ),
    ]

    # declare the same defaults as above but with types
    # to stop LSP complaints
    brightness: float
    brightness_step: float
    disabled_txt: str
    enabled_txt: str
    gamma_red: float
    gamma_blue: float
    gamma_green: float
    redshift_path: str
    temperature: float
    temperature_step: float
    temperature_fmt: str
    brightness_fmt: str
    gamma_fmt: str

    supported_backends = {"x11"}

    def __init__(self, **config):
        _TextBox.__init__(self, **config)
        self.add_defaults(Redshift.defaults)
        self.is_enabled = False
        self._line_index = 0
        self._lines = []
        self.brightness_idx = 1
        self.temperature_idx = 2
        self.gamma_idx = 3
        # redshift's limits
        self.brightness_lower_lim = 0.1
        self.brightness_upper_lim = 1.0
        self.temperature_lower_lim = 1000
        self.temperature_upper_lim = 25000
        self.gamma_lower_lim = 0.1
        self.gamma_upper_lim = 10.0
        # Make sure fields are initialised to values
        # in bounds
        self._assert_brightness()
        self._assert_temperature()
        self.gamma_val = self._assert_gamma()
        self.add_callbacks(
            {
                "Button1": self.click,
                "Button3": self.right_click,
                "Button4": self.scroll_up,
                "Button5": self.scroll_down,
            }
        )
        self.error = None

    def _configure(self, qtile, bar):
        _TextBox._configure(self, qtile, bar)
        # disable redshift so we have a known initial state
        self.reset_redshift()
        # set text after bar configuration is done
        self.qtile.call_soon(self._set_text)

    @expose_command
    def scroll_up(self):
        """
        Scroll up to next item.
        """
        self._scroll(1)

    @expose_command
    def scroll_down(self):
        """
        Scroll down to next item.
        """
        self._scroll(-1)

    def show_line(self):
        """
        Update the text with the the current index in lines.
        """
        assert self._lines, "lines arr should have been initialised"

        line = self._lines[self._line_index]

        self.update(line)

    @expose_command
    def click(self):
        """
        Has no action for the gamma line.
        Either enable or disable the widget if we are currently on the
        first line.
        Else decrease brightness/temperature accordingly
        """
        if self.error or self._line_index == self.gamma_idx:
            return
        elif self._line_index == self.brightness_idx:
            self.decrease_brightness()
        elif self._line_index == self.temperature_idx:
            self.decrease_temperature()
        else:
            self.reset_redshift() if self.is_enabled else self.run_redshift()
            self.is_enabled = not self.is_enabled
        self._set_text()

    @expose_command
    def right_click(self):
        """
        Has no action for the first line of the widget nor the gamma line.
        Increase brightness/temperature accordingly.
        """
        if self.error or self._line_index == 0 or self._line_index == self.gamma_idx:
            return
        elif self._line_index == self.brightness_idx:
            self.increase_brightness()
        elif self._line_index == self.temperature_idx:
            self.increase_temperature()
        self._set_text()

    def _scroll(self, step):
        """
        Scroll up/down dictated by 'step' the items
        and then display that item.

        Has no effect if the widget is disabled.
        """
        if self.error or not self.is_enabled:
            return
        self._line_index = (self._line_index + step) % len(self._lines)
        self.show_line()

    @expose_command
    def decrease_brightness(self):
        """
        Decrease brightness by a single step.
        """
        step = self.brightness_step * -1
        self._change_brightness(step)

    @expose_command
    def increase_brightness(self):
        """
        Increase brightness by a single step.
        """
        self._change_brightness(self.brightness_step)

    def _change_brightness(self, step):
        """
        Control the change brightness logic for both
        decreasing and increasing.
        """
        # To limit the text displayed from float calcs
        self.brightness = round(self.brightness + step, 2)

        # ensure that brightness stays in the valid bounds
        if self.brightness < self.brightness_lower_lim:
            self.brightness = self.brightness_lower_lim
        elif self.brightness > self.brightness_upper_lim:
            self.brightness = self.brightness_upper_lim

        self.run_redshift()

    @expose_command
    def decrease_temperature(self):
        """
        Decrease redshift temperature by a single step.
        """
        step = self.temperature_step * -1
        self._change_temperature(step)

    @expose_command
    def increase_temperature(self):
        """
        Increase redshift temperature by a single step.
        """
        self._change_temperature(self.temperature_step)

    def _change_temperature(self, step):
        """
        Control the change temperature logic for both
        decreasing and increasing.
        """
        # To limit the text displayed from float calcs
        self.temperature = round(self.temperature + step, 2)

        # ensure that temperature stays in the valid bounds
        if self.temperature < self.temperature_lower_lim:
            self.temperature = self.temperature_lower_lim
        elif self.temperature > self.temperature_upper_lim:
            self.temperature = self.temperature_upper_lim

        self.run_redshift()

    def _set_lines(self):
        """
        Set the list of formatted text items to scroll through.
        """
        first_text = ""
        if self.is_enabled:
            first_text = self.enabled_txt
        else:
            first_text = self.disabled_txt

        first_text = self._format_first_text(first_text)

        self._lines = [first_text, "", "", ""]
        self._lines[self.brightness_idx] = self._format_brightness_text()
        self._lines[self.temperature_idx] = self._format_temp_text()
        self._lines[self.gamma_idx] = self._format_gamma_text()

    def _set_text(self):
        """
        Update the lines array and the widget text.
        """
        # Update in case values have changed
        if self.error:
            return
        self._set_lines()
        text = ""
        if self._line_index == 0:
            if self.is_enabled:
                text = self._format_first_text(self.enabled_txt)
            else:
                text = self._format_first_text(self.disabled_txt)
        else:
            text = self._lines[self._line_index]
        self.update(text)

    def _format_temp_text(self) -> str:
        """
        Format the temperature text.
        """
        return self.temperature_fmt.format(temperature=self.temperature)

    def _format_brightness_text(self) -> str:
        """
        Format the brightness text.
        """
        return self.brightness_fmt.format(brightness=self.brightness)

    def _format_gamma_text(self) -> str:
        """
        Format the gamma text.
        """
        return self.gamma_val._format_gamma(self.gamma_fmt)

    def _format_first_text(self, txt: str) -> str:
        """
        Format the enabled/disabled text (aka first_text).
        """
        return txt.format(
            brightness=self.brightness,
            temperature=self.temperature,
            gamma_red=self.gamma_red,
            gamma_green=self.gamma_green,
            gamma_blue=self.gamma_blue,
            is_enabled=self.is_enabled,
        )

    def _assert_brightness(self):
        """
        assert that self.brightness is within
        the correct limits
        """
        assert (
            self.brightness >= self.brightness_lower_lim
            and self.brightness <= self.brightness_upper_lim
        ), (
            f"redshift: self.brightness is not initialised within the acceptable range, see the widget defaults docs: {self.brightness}"
        )

    def _assert_gamma(self) -> GammaGroup:
        """
        assert that self.gamma is within
        the correct limits. If it is, produce
        a GammaGroup instance
        """
        gamma_vals = [
            self.gamma_red,
            self.gamma_green,
            self.gamma_blue,
        ]
        gamma_group = GammaGroup(self.gamma_red, self.gamma_green, self.gamma_blue)
        for _, val in enumerate(gamma_vals):
            assert val >= self.gamma_lower_lim and val <= self.gamma_upper_lim, (
                f"redshift: self.gamma_red, self.gamma_green or self.gamma_blue have not been initialised within the acceptable range, see the docs: {gamma_group}"
            )

        return gamma_group

    def _assert_temperature(self):
        """
        assert that self.temperature is within
        the correct limits
        """
        assert (
            self.temperature >= self.temperature_lower_lim
            and self.temperature <= self.temperature_upper_lim
        ), (
            f"redshift: self.temperature is not initialised within the acceptable range, see the docs: {self.temperature}"
        )

    @expose_command
    def reset_redshift(self):
        """
        Call reset on redshift to reset to default settings.
        """
        try:
            subprocess.run([self.redshift_path, "-x"], check=True)
        except (TypeError, FileNotFoundError) as e:
            self.widget_error(
                f"redshift: could not find redshift executable, check redshift_path: {e}"
            )
        except subprocess.CalledProcessError as e:
            self.widget_error(f"redshift: could not enable redshift: {e}")

    @expose_command
    def run_redshift(self):
        """
        Run redshift command with defined parameters.
        """
        try:
            subprocess.run(
                [
                    self.redshift_path,
                    "-P",
                    "-O",
                    str(self.temperature),
                    "-b",
                    str(self.brightness),
                    "-g",
                    self.gamma_val._redshift_fmt(),
                ],
                check=True,
            )
        except (TypeError, FileNotFoundError) as e:
            self.widget_error(
                f"redshift: could not find redshift executable, check redshift_path: {e}"
            )
        except subprocess.CalledProcessError as e:
            self.widget_error(f"redshift: could not enable redshift: {e}")

    def widget_error(self, error_msg: str):
        """
        Cause the widget to display an error
        and log the given error message
        """
        self.error = error_msg
        logger.exception(self.error)
        self.update("Redshift widget crashed!")



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile import hook
from libqtile.backend.base import Window
from libqtile.scratchpad import ScratchPad

if TYPE_CHECKING:
    from libqtile.backend.base import Static
    from libqtile.core.manager import Qtile


class QtileState:
    """Represents the state of the Qtile object

    This is used for restoring state across restarts or config reloads.

    If `restart` is True, the current set of groups will be saved in the state. This is
    useful when restarting for Qtile version updates rather than reloading the config.
    ScratchPad groups are saved for both reloading and restarting.
    """

    def __init__(self, qtile: Qtile, restart: bool = True) -> None:
        self.groups = []
        self.screens = {}
        self.current_screen = 0
        self.scratchpads = {}
        self.orphans: list[int] = []
        self.restart = restart  # True when restarting, False when config reloading

        for group in qtile.groups:
            if isinstance(group, ScratchPad):
                self.scratchpads[group.name] = group.get_state()
            elif restart:
                self.groups.append((group.name, group.layout.name, group.label))

        for index, screen in enumerate(qtile.screens):
            self.screens[index] = screen.group.name
            if screen == qtile.current_screen:
                self.current_screen = index

    def apply(self, qtile: Qtile) -> None:
        """
        Rearrange the windows in the specified Qtile object according to this
        QtileState.
        """
        for group, layout, label in self.groups:
            try:
                qtile.groups_map[group].layout = layout
                qtile.groups_map[group].label = label
            except KeyError:
                qtile.add_group(group, layout, label=label)

        for screen, group in self.screens.items():
            try:
                group = qtile.groups_map[group]
                qtile.screens[screen].set_group(group)
            except (KeyError, IndexError):
                pass  # group or screen missing

        for group in qtile.groups:
            if isinstance(group, ScratchPad) and group.name in self.scratchpads:
                orphans = group.restore_state(self.scratchpads.pop(group.name), self.restart)
                self.orphans.extend(orphans)
        for sp_state in self.scratchpads.values():
            for _, wid, _ in sp_state:
                self.orphans.append(wid)
        if self.orphans:
            if self.restart:
                hook.subscribe.client_new(self.handle_orphan_dropdowns)
            else:
                for wid in self.orphans:
                    win = qtile.windows_map[wid]
                    if isinstance(win, Window):
                        win.group = qtile.current_group

        qtile.focus_screen(self.current_screen)

    def handle_orphan_dropdowns(self, client: Window | Static) -> None:
        """
        Remove any windows from now non-existent scratchpad groups.
        """
        client_wid = client.wid
        if client_wid in self.orphans:
            self.orphans.remove(client_wid)
            if isinstance(client, Window):
                client.group = None
            if not self.orphans:
                hook.unsubscribe.client_new(self.handle_orphan_dropdowns)



import psutil

from libqtile.widget import base


class CPU(base.InLoopPollText):
    """
    A simple widget to display CPU load and frequency.

    Widget requirements: psutil_.

    .. _psutil: https://pypi.org/project/psutil/
    """

    defaults = [
        ("update_interval", 1.0, "Update interval for the CPU widget"),
        (
            "format",
            "CPU {freq_current}GHz {load_percent}%",
            "CPU display format",
        ),
    ]

    def __init__(self, **config):
        super().__init__("", **config)
        self.add_defaults(CPU.defaults)

    def poll(self):
        variables = dict()

        variables["load_percent"] = round(psutil.cpu_percent(), 1)
        freq = psutil.cpu_freq()
        if psutil.__version__ == "5.9.0":
            variables["freq_current"] = round(freq.current, 1)
        else:
            variables["freq_current"] = round(freq.current / 1000, 1)
        variables["freq_max"] = round(freq.max / 1000, 1)
        variables["freq_min"] = round(freq.min / 1000, 1)

        return self.format.format(**variables)



import os

from libqtile import bar
from libqtile.command.base import expose_command
from libqtile.images import Img
from libqtile.log_utils import logger
from libqtile.widget import base


class Image(base._Widget, base.MarginMixin):
    """Display a PNG image on the bar"""

    orientations = base.ORIENTATION_BOTH
    defaults = [
        ("scale", True, "Enable/Disable image scaling"),
        ("rotate", 0.0, "rotate the image in degrees counter-clockwise"),
        ("filename", None, "Image filename. Can contain '~'"),
        ("margin", 0, "Margin inside the box. Defaults to 0."),
    ]

    def __init__(self, length=bar.CALCULATED, **config):
        base._Widget.__init__(self, length, **config)
        self.add_defaults(Image.defaults)

    def _configure(self, qtile, bar):
        base._Widget._configure(self, qtile, bar)
        self._update_image()

    def _update_image(self):
        self.img = None

        if not self.filename:
            logger.warning("Image filename not set!")
            return

        self.filename = os.path.expanduser(self.filename)

        if not os.path.exists(self.filename):
            logger.warning("Image does not exist: %s", self.filename)
            return

        img = Img.from_path(self.filename)
        self.img = img
        img.theta = self.rotate
        if not self.scale:
            return
        new_height = self.bar.size - (self.margin_top * 2)
        img.resize(height=new_height)

    def draw(self):
        if self.img is None:
            return

        self.drawer.clear(self.background or self.bar.background)
        self.drawer.ctx.save()
        self.drawer.ctx.translate(self.margin_x, self.margin_y)
        self.drawer.ctx.set_source(self.img.pattern)
        self.drawer.ctx.paint()
        self.drawer.ctx.restore()
        self.draw_at_default_position()

    def calculate_length(self):
        if self.img is None:
            return 0
        img_length = self.img.width if self.bar.horizontal else self.img.height
        return img_length + (self.margin_side * 2)

    @expose_command()
    def update(self, filename):
        old_length = self.calculate_length()
        self.filename = filename
        self._update_image()

        if self.calculate_length() == old_length:
            self.draw()
        else:
            self.bar.draw()



from typing import TYPE_CHECKING

from libqtile import bar
from libqtile.widget import base
from libqtile.widget.helpers.status_notifier import has_xdg, host

if TYPE_CHECKING:
    from libqtile.widget.helpers.status_notifier import StatusNotifierItem


class StatusNotifier(base._Widget):
    """
    A 'system tray' widget using the freedesktop StatusNotifierItem
    specification.

    As per the specification, app icons are first retrieved from the
    user's current theme. If this is not available then the app may
    provide its own icon. In order to use this functionality, users
    are recommended to install the `pyxdg <https://pypi.org/project/pyxdg/>`__
    module to support retrieving icons from the selected theme.
    If the icon specified by StatusNotifierItem can not be found in
    the user's current theme and no other icons are provided by the
    app, a fallback icon is used.

    Left-clicking an icon will trigger an activate event.

    .. note::

        Context menus are not currently supported by the official widget.
        However, a modded version of the widget which provides basic menu
        support is available from elParaguayo's `qtile-extras
        <https://github.com/elParaguayo/qtile-extras>`_ repo.
    """

    orientations = base.ORIENTATION_BOTH

    defaults = [
        ("icon_size", 16, "Icon width"),
        ("icon_theme", None, "Name of theme to use for app icons"),
        ("padding", 3, "Padding between icons"),
    ]

    def __init__(self, **config):
        base._Widget.__init__(self, bar.CALCULATED, **config)
        self.add_defaults(StatusNotifier.defaults)
        self.add_callbacks(
            {
                "Button1": self.activate,
            }
        )
        self.selected_item: StatusNotifierItem | None = None

    @property
    def available_icons(self):
        return [item for item in host.items if item.has_icons]

    def calculate_length(self):
        if not host.items:
            return 0

        return len(self.available_icons) * (self.icon_size + self.padding) + self.padding

    def _configure(self, qtile, bar):
        if has_xdg and self.icon_theme:
            host.icon_theme = self.icon_theme

        # This is called last as it starts timers including _config_async.
        base._Widget._configure(self, qtile, bar)

    def draw_callback(self, x=None):
        self.bar.draw()

    async def _config_async(self):
        await host.start(
            on_item_added=self.draw_callback,
            on_item_removed=self.draw_callback,
            on_icon_changed=self.draw_callback,
        )

    def find_icon_at_pos(self, x, y):
        """returns StatusNotifierItem object for icon in given position"""
        offset = self.padding
        val = x if self.bar.horizontal else y

        if val < offset:
            return None

        for icon in self.available_icons:
            offset += self.icon_size
            if val < offset:
                return icon
            offset += self.padding

        return None

    def button_press(self, x, y, button):
        icon = self.find_icon_at_pos(x, y)
        self.selected_item = icon if icon else None

        name = f"Button{button}"
        if name in self.mouse_callbacks:
            self.mouse_callbacks[name]()

    def _draw_icon(self, icon, x, y):
        # Despite scaling the icon down here for compositing, cairo keeps the higher
        # res snapshot, which will be used when the whole buffer is scaled up later
        scale = 1 / getattr(self.bar.window, "scale", 1)
        self.drawer.ctx.save()
        self.drawer.ctx.translate(x, y)
        self.drawer.ctx.scale(scale, scale)
        self.drawer.ctx.set_source_surface(icon, 0, 0)
        self.drawer.ctx.paint()
        self.drawer.ctx.restore()

    def draw(self):
        self.drawer.clear(self.background or self.bar.background)
        xoffset = self.padding
        yoffset = (self.bar.size - self.icon_size) // 2

        # Scale icon up by output scale factor
        scaled_icon_size = int(self.icon_size * getattr(self.bar.window, "scale", 1))
        for item in self.available_icons:
            icon = item.get_icon(scaled_icon_size)
            if self.bar.horizontal:
                self._draw_icon(icon, xoffset, yoffset)
            else:
                self._draw_icon(icon, yoffset, xoffset)
            xoffset += self.icon_size + self.padding

        self.draw_at_default_position()

    def activate(self):
        """Primary action when clicking on an icon"""
        if not self.selected_item:
            return
        self.selected_item.activate()

    def finalize(self):
        host.unregister_callbacks(
            on_item_added=self.draw_callback,
            on_item_removed=self.draw_callback,
            on_icon_changed=self.draw_callback,
        )
        base._Widget.finalize(self)



from __future__ import annotations

import asyncio
import glob
import importlib
import os
from collections import defaultdict
from collections.abc import Sequence
from importlib.metadata import PackageNotFoundError, distribution
from pathlib import Path
from random import randint
from shutil import which
from typing import TYPE_CHECKING, cast

try:
    from dbus_fast import AuthError, Message, Variant
    from dbus_fast.aio import MessageBus
    from dbus_fast.constants import BusType, MessageType

    has_dbus = True
except ImportError:
    has_dbus = False

import libqtile
from libqtile.log_utils import logger

ColorType = str | tuple[int, int, int] | tuple[int, int, int, float]
ColorsType = ColorType | list[ColorType]
if TYPE_CHECKING:
    from collections.abc import Callable, Coroutine
    from typing import Any, TypeVar

    T = TypeVar("T")

try:
    VERSION = distribution("qtile").version
except PackageNotFoundError:
    VERSION = "dev"

dbus_bus_connections = set()

# Create a list to collect references to tasks so they're not garbage collected
# before they've run
TASKS: set[asyncio.Task[None]] = set()


def create_task(coro: Coroutine) -> asyncio.Task | None:
    """
    Wrapper for asyncio.create_task.

    Stores task so garbage collector doesn't remove it and removes reference when it's done.
    See: https://textual.textualize.io/blog/2023/02/11/the-heisenbug-lurking-in-your-async-code/
    for more info about the issue this solves.
    """
    loop = asyncio.get_running_loop()
    if not loop:
        return None

    task = asyncio.create_task(coro)
    TASKS.add(task)
    task.add_done_callback(TASKS.discard)

    return task


class QtileError(Exception):
    pass


def lget[T](o: list[T], v: int) -> T | None:
    try:
        return o[v]
    except (IndexError, TypeError):
        return None


def rgb(x: ColorType) -> tuple[float, float, float, float]:
    """
    Returns a valid RGBA tuple.

    Here are some valid specifications:
        #ff0000
        with alpha: #ff000080
        ff0000
        with alpha: ff0000.5
        (255, 0, 0)
        with alpha: (255, 0, 0, 0.5)

    Which is returned as (1.0, 0.0, 0.0, 0.5).
    """
    if isinstance(x, tuple | list):
        if len(x) == 4:
            alpha = x[-1]
        else:
            alpha = 1.0
        return (x[0] / 255.0, x[1] / 255.0, x[2] / 255.0, alpha)
    elif isinstance(x, str):
        if x.startswith("#"):
            x = x[1:]
        if "." in x:
            x, alpha_str = x.split(".")
            alpha = float("0." + alpha_str)
        else:
            alpha = 1.0
        if len(x) not in (3, 6, 8):
            raise ValueError("RGB specifier must be 3, 6 or 8 characters long.")
        if len(x) == 3:
            # Multiplying by 17: 0xA * 17 = 0xAA etc.
            vals = tuple(int(i, 16) * 17 for i in x)
        else:
            vals = tuple(int(i, 16) for i in (x[0:2], x[2:4], x[4:6]))
        if len(x) == 8:
            alpha = int(x[6:8], 16) / 255.0
        vals += (alpha,)  # type: ignore
        return rgb(vals)  # type: ignore
    raise ValueError("Invalid RGB specifier.")


def hex(x: ColorType) -> str:
    r, g, b, _ = rgb(x)
    return f"#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}"


def has_transparency(colour: ColorsType) -> bool:
    """
    Returns True if the colour is not fully opaque.

    Where a list of colours is passed, returns True if any
    colour is not fully opaque.
    """
    if isinstance(colour, str | tuple):
        return rgb(colour)[3] < 1
    return any(has_transparency(c) for c in colour)


def remove_transparency(colour: ColorsType) -> ColorsType:
    """
    Returns a tuple of (r, g, b) with no alpha.
    """
    if isinstance(colour, str | tuple):
        r, g, b, _ = rgb(colour)
        return (int(r * 255), int(g * 255), int(b * 255))
    return [cast(ColorType, remove_transparency(c)) for c in colour]


def is_valid_colors(color: ColorsType) -> bool:
    """
    Returns whether the argument is a valid color or list of colors.
    """
    if not isinstance(color, list):
        color = [color]
    try:
        list(rgb(c) for c in color)
        return True
    except (ValueError, TypeError):
        return False


def scrub_to_utf8(text: str | bytes) -> str:
    if not text:
        return ""
    elif isinstance(text, str):
        return text
    else:
        return text.decode("utf-8", "ignore")


def get_cache_dir() -> str:
    """
    Returns the cache directory and create if it doesn't exists
    """

    cache_directory = os.path.expandvars("$XDG_CACHE_HOME")
    if cache_directory == "$XDG_CACHE_HOME":
        # if variable wasn't set
        cache_directory = os.path.expanduser("~/.cache")
    cache_directory = os.path.join(cache_directory, "qtile")
    if not os.path.exists(cache_directory):
        os.makedirs(cache_directory)
    return cache_directory


def get_config_file() -> Path:
    config_home = Path(os.getenv("XDG_CONFIG_HOME", "~/.config")).expanduser()
    config_file = config_home.joinpath("qtile/config.py")
    if config_file.exists():
        return config_file

    xdg_config_dirs = os.getenv("XDG_CONFIG_DIRS", "/etc/xdg/").split(":")
    for config_dir in xdg_config_dirs:
        system_wide_config = Path(config_dir).expanduser().joinpath("qtile/config.py")
        if system_wide_config.exists():
            return system_wide_config

    return config_file


def describe_attributes(obj: Any, attrs: list[str], func: Callable = lambda x: x) -> str:
    """
    Helper for __repr__ functions to list attributes with truthy values only
    (or values that return a truthy value by func)
    """

    pairs = []

    for attr in attrs:
        value = getattr(obj, attr, None)
        if func(value):
            pairs.append(f"{attr}={value}")

    return ", ".join(pairs)


def import_class(
    module_path: str,
    class_name: str,
    fallback: Callable | None = None,
) -> Any:
    """Import a class safely

    Try to import the class module, and if it fails because of an ImportError
    it logs on WARNING, and logs the traceback on DEBUG level
    """
    try:
        module = importlib.import_module(module_path, __package__)
        return getattr(module, class_name)
    except ImportError:
        if fallback:
            logger.error(
                "Unmet dependencies for '%s.%s'. Using fallback.", module_path, class_name
            )
            return fallback(module_path, class_name)

        logger.exception("Unmet dependencies for '%s.%s':", module_path, class_name)
        raise


def lazify_imports(
    registry: dict[str, str],
    package: str,
    fallback: Callable | None = None,
) -> tuple[tuple[str, ...], Callable, Callable]:
    """Leverage PEP 562 to make imports lazy in an __init__.py

    The registry must be a dictionary with the items to import as keys and the
    modules they belong to as a value.
    """
    __all__ = tuple(registry.keys())

    def __dir__() -> tuple[str, ...]:  # noqa: N807
        return __all__

    def __getattr__(name: str) -> Any:  # noqa: N807
        if name not in registry:
            raise AttributeError
        module_path = f"{package}.{registry[name]}"
        return import_class(module_path, name, fallback=fallback)

    return __all__, __dir__, __getattr__


def send_notification(
    title: str,
    message: str,
    urgent: bool = False,
    timeout: int = -1,
    id_: int | None = None,
) -> int:
    """
    Send a notification.

    The id_ argument, if passed, requests the notification server to replace a visible
    notification with the same ID. An ID is returned for each call; this would then be
    passed when calling this function again to replace that notification. See:
    https://developer.gnome.org/notification-spec/
    """
    if "PYTEST_CURRENT_TEST" in os.environ:
        logger.warning("skipped notification because we are in tests")
        return -1
    if not has_dbus:
        logger.warning("dbus-fast is not installed. Unable to send notifications.")
        return -1

    id_ = randint(10, 1000) if id_ is None else id_
    urgency = 2 if urgent else 1

    loop = libqtile.event_loop
    loop.create_task(_notify(title, message, urgency, timeout, id_))

    return id_


async def _notify(
    title: str,
    message: str,
    urgency: int,
    timeout: int,
    id_: int,
) -> None:
    notification = [
        "qtile",  # Application name
        id_,  # id
        "",  # icon
        title,  # summary
        message,  # body
        [],  # actions
        {"urgency": Variant("y", urgency)},  # hints
        timeout,
    ]  # timeout

    bus, msg = await _send_dbus_message(
        True,
        MessageType.METHOD_CALL,
        "org.freedesktop.Notifications",
        "org.freedesktop.Notifications",
        "/org/freedesktop/Notifications",
        "Notify",
        "susssasa{sv}i",
        notification,
    )

    if msg and msg.message_type == MessageType.ERROR:
        logger.warning("Unable to send notification. Is a notification server running?")

    # a new bus connection is made each time a notification is sent so
    # we disconnect when the notification is done
    if bus:
        bus.disconnect()


def guess_terminal(preference: str | Sequence | None = None) -> str | None:
    """Try to guess terminal."""
    test_terminals = []
    if isinstance(preference, str):
        test_terminals += [preference]
    elif isinstance(preference, Sequence):
        test_terminals += list(preference)
    if "WAYLAND_DISPLAY" in os.environ:
        # Wayland-only terminals
        test_terminals += [
            "foot",
            "mlterm-wl",
        ]
    test_terminals += [
        "alacritty",
        "cool-retro-term",
        "coreterminal",
        "eterm",
        "ghostty",
        "gnome-terminal",
        "guake",
        "hyper",
        "kitty",
        "konsole",
        "lxterminal",
        "mate-terminal",
        "mlterm",
        "ptyxis",
        "qterminal",
        "roxterm",
        "sakura",
        "st",
        "terminator",
        "terminology",
        "tilda",
        "urxvt",
        "wezterm",
        "x-terminal-emulator",
        "xfce4-terminal",
        "xterm",
        "yakuake",
        "zutty",
    ]

    for terminal in test_terminals:
        logger.debug("Guessing terminal: %s", terminal)
        if not which(terminal, os.X_OK):
            continue

        logger.info("Terminal found: %s", terminal)
        return terminal

    logger.error("Default terminal has not been found.")
    return None


def scan_files(dirpath: str, *names: str) -> defaultdict[str, list[str]]:
    """
    Search a folder recursively for files matching those passed as arguments, with
    globbing. Returns a dict with keys equal to entries in names, and values a list of
    matching paths. E.g.:

    >>> scan_files('/wallpapers', '*.png', '*.jpg')
    defaultdict(<class 'list'>, {'*.png': ['/wallpapers/w1.png'], '*.jpg':
    ['/wallpapers/w2.jpg', '/wallpapers/w3.jpg']})

    """
    dirpath = os.path.expanduser(dirpath)
    files = defaultdict(list)

    for name in names:
        found = glob.glob(os.path.join(dirpath, "**", name), recursive=True)
        files[name].extend(found)

    return files


async def _send_dbus_message(
    session_bus: bool,
    message_type: MessageType,
    destination: str | None,
    interface: str | None,
    path: str | None,
    member: str | None,
    signature: str,
    body: Any,
    negotiate_unix_fd: bool = False,
    bus: MessageBus | None = None,
    preserve: bool = False,
) -> tuple[MessageBus | None, Message | None]:
    """
    Private method to send messages to dbus via dbus_fast.

    An existing bus connection can be passed, if left empty, a new
    bus connection will be created.

    Returns a tuple of the bus object and message response.
    """
    if bus is None:
        if session_bus:
            bus_type = BusType.SESSION
        else:
            bus_type = BusType.SYSTEM

        try:
            bus = await MessageBus(
                bus_type=bus_type, negotiate_unix_fd=negotiate_unix_fd
            ).connect()
        except (AuthError, Exception):
            logger.warning("Unable to connect to dbus.")
            return None, None

    if isinstance(body, str):
        body = [body]

    # Ignore types here: dbus-fast has default values of `None` for certain
    # parameters but the signature is `str` so passing `None` results in an
    # error in mypy.
    msg = await bus.call(
        Message(
            message_type=message_type,
            destination=destination,
            interface=interface,
            path=path,
            member=member,
            signature=signature,
            body=body,
        )
    )

    # Keep details of bus connections so we can close them on exit
    # dbus_bus_connetions is a set so we don't need to worry about
    # duplicates
    if not preserve:
        dbus_bus_connections.add(bus)

    return bus, msg


async def add_signal_receiver(
    callback: Callable,
    session_bus: bool = False,
    signal_name: str | None = None,
    dbus_interface: str | None = None,
    bus_name: str | None = None,
    path: str | None = None,
    check_service: bool = False,
    use_bus: MessageBus | None = None,
    preserve: bool = False,
) -> bool:
    """
    Helper function which aims to recreate python-dbus's add_signal_receiver
    method in dbus_fast with asyncio calls.

    If check_service is `True` the method will raise a wanrning and return False
    if the service is not visible on the bus. If the `bus_name` is None, no
    check will be performed.

    Returns True if subscription is successful.
    """
    if not has_dbus:
        logger.warning("dbus-fast is not installed. Unable to subscribe to signals")
        return False

    if bus_name and check_service:
        found = await find_dbus_service(bus_name, session_bus)
        if not found:
            logger.warning(
                "The %s name was not found on the bus. No callback will be attached.", bus_name
            )
            return False

    match_args = {
        "sender": bus_name,
        "member": signal_name,
        "path": path,
        "interface": dbus_interface,
    }

    rule = "type='signal',"
    rule += ",".join(f"{k}='{v}'" for k, v in match_args.items() if v)

    logger.debug("Adding dbus match rule: %s", rule)

    bus, msg = await _send_dbus_message(
        session_bus,
        MessageType.METHOD_CALL,
        "org.freedesktop.DBus",
        "org.freedesktop.DBus",
        "/org/freedesktop/DBus",
        "AddMatch",
        "s",
        [rule],
        bus=use_bus,
        preserve=preserve,
    )

    # Check if message sent successfully
    if bus and msg and msg.message_type == MessageType.METHOD_RETURN:

        def match_message(msg: Message, match_args: dict[str, str | None]) -> bool:
            return all(getattr(msg, k) == v for k, v in match_args.items() if v)

        async def resolve_sender(signal_msg: Message) -> tuple[str, Message]:
            """Looks up a pretty bus name to retrieve the unique name."""
            _, sender_msg = await _send_dbus_message(
                session_bus,
                MessageType.METHOD_CALL,
                "org.freedesktop.DBus",
                "org.freedesktop.DBus",
                "/org/freedesktop/DBus",
                "GetNameOwner",
                "s",
                [match_args["sender"]],
                bus=bus,
            )

            if sender_msg and sender_msg.message_type == MessageType.METHOD_RETURN:
                return sender_msg.body[0], signal_msg

            return "", signal_msg

        def check_message(task: asyncio.Task) -> None:
            new_match_args = match_args.copy()
            new_sender, signal_message = task.result()
            new_match_args["sender"] = new_sender
            if match_message(signal_message, new_match_args):
                callback(signal_message)

        def signal_callback_wrapper(msg: Message) -> None:
            """Custom wrapper to only run callback if message matches our rule."""
            if msg.message_type == MessageType.SIGNAL:
                if match_message(msg, match_args):
                    callback(msg)
                elif "sender" in match_args:
                    # If the message didn't match and we're trying to match the sender
                    # We may need to convert the pretty name to the bus's unique name first
                    task = create_task(resolve_sender(msg))
                    if task:
                        task.add_done_callback(check_message)

        bus.add_message_handler(signal_callback_wrapper)
        return True

    else:
        return False


async def find_dbus_service(service: str, session_bus: bool) -> bool:
    """Looks up service name to see if it is currently available on dbus."""

    # We're using low level interface here to reduce unnecessary calls for
    # introspection etc.
    bus, msg = await _send_dbus_message(
        session_bus,
        MessageType.METHOD_CALL,
        "org.freedesktop.DBus",
        "org.freedesktop.DBus",
        "/org/freedesktop/DBus",
        "ListNames",
        "",
        [],
    )

    if bus is None or msg is None or (msg and msg.message_type != MessageType.METHOD_RETURN):
        logger.warning("Unable to send lookup call to dbus.")
        return False

    bus.disconnect()

    names = msg.body[0]

    return service in names


def remove_dbus_rules() -> None:
    # Disconnecting the bus connections is enough to remove the match rules.
    while dbus_bus_connections:
        bus = dbus_bus_connections.pop()
        try:
            bus.disconnect()
        except OSError:
            # Socket has already shut down
            pass

        # We need to manually close the socket until https://github.com/altdesktop/python-dbus-next/pull/148
        # gets merged. There's no error on multiple calls to 'close()'.
        if bus._sock is not None:
            bus._sock.close()


ASYNC_PIDS: set[int] = set()


async def acall_process(command: str | list[str], shell: bool = False) -> str:
    """
    Like call_process, but the async version. Tracks PIDs in ASYNC_PIDS.
    """
    stdin = asyncio.subprocess.DEVNULL
    stdout = asyncio.subprocess.PIPE
    stderr = asyncio.subprocess.STDOUT

    if shell:
        if isinstance(command, list):
            command = " ".join(command)
        p = await asyncio.subprocess.create_subprocess_shell(
            command, stdin=stdin, stdout=stdout, stderr=stderr
        )
    else:
        if isinstance(command, str):
            command = [command]
        p = await asyncio.subprocess.create_subprocess_exec(
            *command, stdin=stdin, stdout=stdout, stderr=stderr
        )

    if p.pid is not None:
        ASYNC_PIDS.add(p.pid)

    try:
        (out, _) = await p.communicate()
        return out.decode("utf-8")
    finally:
        # Remove PID from tracking list when process completes
        if p.pid is not None:
            ASYNC_PIDS.discard(p.pid)


def reap_zombies() -> None:
    """
    A SIGCHLD handler that reaps all zombies until there are no more.
    Uses WNOWAIT for processes in ASYNC_PIDS to avoid interfering with async cleanup.
    """
    try:
        # One signal might mean mulitple children have exited. Reap everything
        # that has exited, until there's nothing left.
        while True:
            wait_result = os.waitid(os.P_ALL, 0, os.WEXITED | os.WNOHANG | os.WNOWAIT)
            if wait_result is None:
                return

            if wait_result.si_pid in ASYNC_PIDS:
                break
            os.waitid(os.P_PID, wait_result.si_pid, os.WEXITED)
    except ChildProcessError:
        pass



from itertools import filterfalse

import libcst as cst
import libcst.matchers as m

from libqtile.scripts.migrations._base import (
    Check,
    MigrationTransformer,
    _QtileMigrator,
    add_migration,
)


class RenameUnspecifiedTransformer(MigrationTransformer):
    @m.leave(m.Call(func=m.Attribute(attr=m.Name("set_font"))))
    def update_unspecified_arg(self, original_node, updated_node) -> cst.Call:
        args = original_node.args

        def is_interesting_arg(arg):
            return m.matches(arg.value, m.Name("UNSPECIFIED"))

        args = list(filterfalse(is_interesting_arg, original_node.args))

        if len(args) == len(original_node.args):
            return original_node

        return updated_node.with_changes(args=args)

    def strip_unspecified_imports(self, original_node, updated_node) -> cst.Import:
        new_names = list(filter(lambda n: n.name.value != "UNSPECIFIED", original_node.names))
        if len(new_names) == len(original_node.names):
            return original_node
        self.lint(original_node, "'UNSPECIFIED' can be dropped")
        if len(new_names) == 0:
            return cst.RemoveFromParent()
        return updated_node.with_changes(names=new_names)

    def leave_Import(self, original_node: cst.Import, updated_node: cst.Import):  # noqa: N802
        return self.strip_unspecified_imports(original_node, updated_node)

    def leave_ImportFrom(  # noqa: N802
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ):
        return self.strip_unspecified_imports(original_node, updated_node)


class RenameUnspecified(_QtileMigrator):
    ID = "RenameUnspecified"

    SUMMARY = "Drops `UNSPECIFIED` argument"

    HELP = """
    The UNSPECIFIED object was removed in favor of using the zero values (or
    None) to leave behavior unspecified. That is:

        font=UNSPECIFIED -> font=None
        fontsize=UNSPECIFIED -> fontsize=0
        fontshadow=UNSPECIFIED -> fontshadow=""
    """

    AFTER_VERSION = "0.25.0"

    TESTS = [
        Check(
            """
            from typing import TYPE_CHECKING
            from libqtile.widget.base import UNSPECIFIED, ORIENTATION_BOTH
            from libqtile.widget import TextBox
            from libqtile.layout import Tile

            tb = TextBox(text="hello")
            # just to use ORIENTATION_BOTH and force us to delete only the
            # right thing
            if TYPE_CHECKING:
                tb.orientations = ORIENTATION_BOTH
                tb.set_font(font=UNSPECIFIED, fontsize=12)
            """,
            """
            from typing import TYPE_CHECKING
            from libqtile.widget.base import ORIENTATION_BOTH
            from libqtile.widget import TextBox
            from libqtile.layout import Tile

            tb = TextBox(text="hello")
            # just to use ORIENTATION_BOTH and force us to delete only the
            # right thing
            if TYPE_CHECKING:
                tb.orientations = ORIENTATION_BOTH
                tb.set_font(fontsize=12)
            """,
        )
    ]

    visitor = RenameUnspecifiedTransformer


add_migration(RenameUnspecified)



from __future__ import annotations

from typing import TYPE_CHECKING

from libqtile.backend.base.idle_inhibit import IdleInhibitorManager as BaseIdleInhibitorManager
from libqtile.backend.base.idle_inhibit import Inhibitor, InhibitorType

try:
    from libqtile.backend.wayland._ffi import ffi, lib
except ModuleNotFoundError:
    from libqtile.backend.wayland.ffi_stub import ffi, lib

if TYPE_CHECKING:
    from typing import Any

    from libqtile.backend.base.window import Window
    from libqtile.core.manager import Qtile


class WaylandInhibitor(Inhibitor):
    def __init__(
        self,
        qtile: Qtile,
        window: Window | None = None,
        handle: ffi.CData | None = None,
        inhibitor_type: InhibitorType = InhibitorType.UNSET,
        is_layer_surface: bool = False,
        is_session_lock: bool = False,
    ):
        super().__init__(qtile, window, inhibitor_type)
        self.handle = handle
        self.is_layer_surface = is_layer_surface
        self.is_session_lock = is_session_lock

    def check(self) -> bool:
        # If a session lock is active we only allow inhibitors from the session lock
        if self.qtile.locked:
            return self.is_session_lock and lib.qw_server_inhibitor_surface_visible(
                self.handle, ffi.NULL
            )

        if self.inhibitor_type == InhibitorType.APPLICATION:
            # Application-set inhibitors should apply when the client is visible
            if self.window is not None:
                active = self.window.is_visible()
            elif self.is_session_lock or self.is_layer_surface:
                active = lib.qw_server_inhibitor_surface_visible(self.handle, ffi.NULL)
            # We shouldn't really get here but, if we do, treat it as active
            else:
                active = True
            return active

        return super().check()

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, WaylandInhibitor):
            raise NotImplementedError

        return (
            self.window,
            self.handle,
            self.inhibitor_type,
            self.is_layer_surface,
            self.is_session_lock,
        ) == (
            other.window,
            other.handle,
            other.inhibitor_type,
            other.is_layer_surface,
            other.is_session_lock,
        )

    @classmethod
    def from_base_inhibitor(cls, inhibitor: Inhibitor) -> WaylandInhibitor:
        if isinstance(inhibitor, WaylandInhibitor):
            return inhibitor

        return cls(
            qtile=inhibitor.qtile,
            window=inhibitor.window,
            inhibitor_type=inhibitor.inhibitor_type,
        )


class IdleInhibitorManager(BaseIdleInhibitorManager[WaylandInhibitor]):
    def _convert_inhibitor(self, inhibitor: Inhibitor) -> WaylandInhibitor:
        return WaylandInhibitor.from_base_inhibitor(inhibitor)

    def add_extension_inhibitor(
        self,
        handle: ffi.CData,
        window: Window | None,
        is_layer_surface: bool,
        is_session_lock: bool,
    ) -> bool:
        inhibitor = WaylandInhibitor(
            qtile=self.core.qtile,
            window=window,
            handle=handle,
            inhibitor_type=InhibitorType.APPLICATION,
            is_layer_surface=is_layer_surface,
            is_session_lock=is_session_lock,
        )
        return self.add_inhibitor_safe(inhibitor)

    def remove_extension_inhibitor(self, pointer: ffi.CData) -> bool:
        return self.remove_inhibitor(lambda o: o.handle == pointer)
